from tempfile import mkdtemp

from django.contrib import messages
from django.utils import timezone
from reportlab.graphics.barcode import code39
from reportlab.lib import colors
from reportlab.lib.units import cm, mm
from reportlab.platypus import Paragraph, Spacer, Table, TableStyle

from edc_appointment.models import Appointment
from edc_constants.constants import YES
from edc_identifier.utils import convert_to_human_readable
from edc_lab.model_mixins import RequisitionModelMixin
from edc_lab.models.manifest.shipper import Shipper
from edc_pdf_reports import Report


class RequisitionReport(Report):
    def __init__(
        self,
        *,
        appointment: Appointment,
        selected_panel_names: list[str],
        consignee=None,
        request=None,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self._requisitions = []
        self.request = request
        self.selected_panel_names = selected_panel_names
        self.site = request.site
        self.appointment = appointment
        self.user = request.user
        self.consignee = consignee
        self.contact_name = f"{self.user.first_name} {self.user.last_name}"
        self.image_folder = mkdtemp()
        self.timestamp = timezone.now().strftime("%Y%m%d%H%M%S")

    @property
    def report_filename(self) -> str:
        return f"requisition_{self.timestamp}.pdf"

    @property
    def shipper(self):
        """Return a shipper model instance."""
        try:
            shipper = Shipper.objects.all()[0]
        except IndexError:
            messages.error(self.request, "Unable to print report. Please define the shipper.")
            shipper = Shipper()
        return shipper

    def get_report_story(self, **kwargs):  # noqa: ARG002
        story = [
            Paragraph(
                "PATIENT SPECIMEN REQUISITION / MANIFEST",
                self.styles["line_label_center"],
            )
        ]

        data = [
            [
                Paragraph("DATE", self.styles["line_label"]),
                Paragraph("REFERENCE", self.styles["line_label"]),
            ],
            [
                Paragraph(
                    timezone.now().strftime("%Y-%m-%d"), self.styles["line_data_largest"]
                ),
                Paragraph(
                    convert_to_human_readable(self.timestamp),
                    self.styles["line_data_largest"],
                ),
            ],
        ]
        t = Table(data)
        t.setStyle(
            TableStyle(
                [
                    ("INNERGRID", (0, 0), (1, 0), 0.25, colors.black),
                    ("INNERGRID", (0, 1), (1, 1), 0.25, colors.black),
                    ("BOX", (0, 0), (-1, -1), 0.25, colors.black),
                    ("VALIGN", (0, 0), (-1, -1), "MIDDLE"),
                ]
            )
        )
        story.append(t)

        data = [
            [
                Paragraph("FROM (complete name and address)", self.styles["line_label"]),
                Paragraph("DELIVER TO (complete name and address)", self.styles["line_label"]),
            ],
            [
                Paragraph(
                    self.formatted_address(**self.shipper_data),
                    self.styles["line_data_large"],
                ),
                Paragraph(
                    self.formatted_address(**self.consignee_data),
                    self.styles["line_data_large"],
                ),
            ],
        ]

        t = Table(data)
        t.setStyle(
            TableStyle(
                [
                    ("INNERGRID", (0, 0), (1, 0), 0.25, colors.black),
                    ("INNERGRID", (0, 1), (1, 1), 0.25, colors.black),
                    ("BOX", (0, 0), (-1, -1), 0.25, colors.black),
                    ("VALIGN", (0, 0), (-1, -1), "MIDDLE"),
                ]
            )
        )
        story.append(t)

        data = [
            [
                Paragraph("CLINIC NAME", self.styles["line_label"]),
                Paragraph("INSTRUCTIONS<br />", self.styles["line_label"]),
            ],
            [
                Paragraph(self.site.name.title(), self.styles["line_data_large"]),
                Paragraph(self.description or "", self.styles["line_data"]),
            ],
            [Paragraph("TEL/MOBILE/FAX", self.styles["line_label"]), ""],
            [
                Paragraph(
                    f"T:{self.shipper.telephone or '?'} M:{self.shipper.mobile or '?'} "
                    f"F:{self.shipper.fax or '?'}",
                    self.styles["line_data_large"],
                ),
                "",
            ],
            [Paragraph("EMAIL", self.styles["line_label"]), ""],
            [Paragraph(self.shipper.email or "", self.styles["line_data_large"]), ""],
        ]
        t1 = Table(data)
        t1.setStyle(
            TableStyle(
                [
                    ("INNERGRID", (0, 1), (0, 2), 0.25, colors.black),
                    ("INNERGRID", (0, 3), (0, 4), 0.25, colors.black),
                    ("INNERGRID", (0, 0), (1, 0), 0.25, colors.black),
                    ("INNERGRID", (0, 1), (1, 1), 0.25, colors.black),
                    ("INNERGRID", (0, 2), (1, 2), 0.25, colors.black),
                    ("INNERGRID", (0, 3), (1, 3), 0.25, colors.black),
                    ("INNERGRID", (0, 4), (1, 4), 0.25, colors.black),
                    ("INNERGRID", (0, 5), (1, 5), 0.25, colors.black),
                    ("SPAN", (1, 1), (1, 5)),
                    ("BOX", (0, 0), (-1, -1), 0.25, colors.black),
                    ("VALIGN", (0, 0), (-1, -1), "MIDDLE"),
                ]
            )
        )
        story.append(t1)

        story.append(Spacer(0.1 * cm, 0.5 * cm))

        story.append(
            Table(
                [
                    [
                        Paragraph(
                            "I DECLARE THE INFORMATION TO BE TRUE AND CORRECT",
                            self.styles["line_label"],
                        )
                    ]
                ]
            )
        )

        story.append(Spacer(0.1 * cm, 0.5 * cm))

        data = [
            [
                Paragraph(self.contact_name, self.styles["line_data_large"]),
                "",
                Paragraph(
                    timezone.now().strftime("%Y-%m-%d %H:%M"),
                    self.styles["line_data_large"],
                ),
            ],
            [
                Paragraph("SIGNATURE (sign next to your name.)", self.styles["line_label"]),
                "",
                Paragraph("DATE", self.styles["line_label"]),
            ],
        ]

        t1 = Table(data, colWidths=(None, 2 * cm, None))
        t1.setStyle(
            TableStyle(
                [
                    ("INNERGRID", (0, 0), (0, 1), 0.25, colors.black),
                    ("INNERGRID", (2, 0), (2, 1), 0.25, colors.black),
                ]
            )
        )

        story.append(t1)
        story.append(Spacer(0.1 * cm, 0.5 * cm))

        story.append(
            Paragraph(
                "Place a check-mark in the right-most column for each specimen received.",
                self.styles["line_label"],
            )
        )

        story = self.append_requisition_items_story(story)

        story.append(Spacer(0.1 * cm, 1 * cm))

        story.append(Paragraph("Comments (if any):", self.styles["line_label"]))
        story.append(Spacer(0.1 * cm, 2 * cm))

        story.append(
            Paragraph(
                "The above items marked with a check have been received in good order:",
                self.styles["line_label"],
            )
        )

        story.append(Spacer(0.1 * cm, 1 * cm))

        data = [
            [
                Paragraph("", self.styles["line_data_large"]),
                "",
                Paragraph("", self.styles["line_data_large"]),
            ],
            [
                Paragraph(
                    "SIGNATURE OF LABORATORY TECHNICIAN (Write name and sign.)",
                    self.styles["line_label"],
                ),
                "",
                Paragraph("DATE", self.styles["line_label"]),
            ],
        ]

        t1 = Table(data, colWidths=(None, 2 * cm, None))
        t1.setStyle(
            TableStyle(
                [
                    ("INNERGRID", (0, 0), (0, 1), 0.25, colors.black),
                    ("INNERGRID", (2, 0), (2, 1), 0.25, colors.black),
                ]
            )
        )
        story.append(t1)

        return story

    @property
    def requisitions(self):
        """Returns a list of requisition model instances related
        to this appointment.
        """
        if not self._requisitions:
            for k, v in self.appointment.related_visit_model_cls().__dict__.items():
                try:
                    model_cls = v.rel.related_model
                except AttributeError:
                    pass
                else:
                    if issubclass(model_cls, RequisitionModelMixin):
                        self._requisitions.extend(
                            list(
                                getattr(self.appointment.related_visit, k).filter(
                                    clinic_verified=YES,
                                    panel__name__in=self.selected_panel_names,
                                )
                            )
                        )
        return self._requisitions

    @property
    def shipper_data(self):
        data = self.shipper.__dict__
        if self.contact_name.strip():
            data.update(contact_name=self.contact_name)
        return data

    @property
    def consignee_data(self):
        return self.consignee.__dict__

    @staticmethod
    def formatted_address(**kwargs):
        data = {
            "contact_name": None,
            "name": None,
            "address": None,
            "city": None,
            "state": None,
            "postal_code": "0000",
            "country": None,
        }
        data.update(**kwargs)
        data_list = [
            v
            for v in [
                data.get("contact_name"),
                data.get("name"),
                data.get("address"),
                (
                    data.get("city")
                    if not data.get("state")
                    else "{} {}".format(data.get("city"), data.get("state"))
                ),
                data.get("postal_code"),
                data.get("country"),
            ]
            if v
        ]
        return "<br />".join(data_list)

    @property
    def description(self):
        return (
            "Please process the following specimens according to the "
            "panels listed below. Use the specimen identifier and "
            "subject identifier as a reference for panel results."
        )

    def append_requisition_items_story(self, story):
        story.append(Spacer(0.1 * cm, 0.5 * cm))
        index = 0
        for requisition in self.requisitions:
            table_data = []
            if index == 0:
                table_data.append(
                    [
                        Paragraph("BARCODE", self.styles["line_label_center"]),
                        Paragraph("ITEM", self.styles["line_label_center"]),
                        Paragraph("SPECIMEN", self.styles["line_label_center"]),
                        Paragraph("SUBJECT", self.styles["line_label_center"]),
                        Paragraph("PANEL", self.styles["line_label_center"]),
                        Paragraph("TYPE", self.styles["line_label_center"]),
                        Paragraph("DRAWN", self.styles["line_label_center"]),
                        Paragraph("RECV'D", self.styles["line_label_center"]),
                    ]
                )
            barcode = code39.Standard39(
                requisition.requisition_identifier, barHeight=10 * mm, stop=1
            )
            item_count = requisition.item_count or 1
            for count in range(1, item_count + 1):
                table_data.append(
                    [
                        barcode,
                        Paragraph(str(index + 1), self.styles["row_data"]),
                        Paragraph(
                            f"{requisition.human_readable_identifier} ({count}/{item_count})",
                            self.styles["row_data"],
                        ),
                        Paragraph(
                            requisition.subject_identifier or "?",
                            self.styles["row_data"],
                        ),
                        Paragraph(
                            f"{requisition.panel_object.verbose_name} <br />"
                            f"({requisition.panel_object.abbreviation})",
                            self.styles["row_data"],
                        ),
                        Paragraph(
                            f"{requisition.panel_object.aliquot_type}",
                            self.styles["row_data"],
                        ),
                        Paragraph(
                            requisition.drawn_datetime.strftime("%Y-%m-%d"),
                            self.styles["row_data"],
                        ),
                        Paragraph("", self.styles["row_data"]),
                    ]
                )
                index += 1

            t1 = Table(table_data)
            t1.setStyle(
                TableStyle(
                    [
                        ("INNERGRID", (0, 0), (-1, -1), 0.25, colors.black),
                        ("VALIGN", (0, 0), (-1, -1), "MIDDLE"),
                        ("BOX", (0, 0), (-1, -1), 0.25, colors.black),
                    ]
                )
            )
            story.append(t1)

        return story



from __future__ import annotations

import sys
from importlib import import_module
from typing import TYPE_CHECKING

from django.apps import apps as django_apps
from django.conf import settings

from ..formula import Formula
from .get_default_reportable_grades import get_default_reportable_grades
from .reference_range_colllection_model_cls import reference_range_colllection_model_cls
from .update_grading_data import update_grading_data
from .update_grading_exceptions import update_grading_exceptions
from .update_normal_data import update_normal_data

if TYPE_CHECKING:
    from ..models import ReferenceRangeCollection


class AlreadyLoaded(Exception):
    pass


__all__ = ["load_all_reference_ranges", "load_reference_ranges"]


def get_module_name() -> str:
    return getattr(settings, "EDC_REPORTABLE_DEFAULT_MODULE_NAME", "reportables")


def load_all_reference_ranges() -> None:
    """Check each app and load the reference ranges if the module
    exists.

    Typically called by post_migrate.
    """
    module_name = get_module_name()
    sys.stdout.write(f" * checking for site {module_name} ...\n")
    for app in django_apps.app_configs:
        try:
            reportables_module = import_module(f"{app}.{module_name}")
        except ImportError:
            pass
        else:
            load_reference_ranges(
                reportables_module.collection_name,
                normal_data=reportables_module.normal_data,
                grading_data=reportables_module.grading_data,
                reportable_grades=reportables_module.reportable_grades,
                reportable_grades_exceptions=reportables_module.reportable_grades_exceptions,
            )
            sys.stdout.write(
                f"   - loaded {app}.{module_name} collection "
                f"`{reportables_module.collection_name}`."
            )


def load_reference_ranges(
    collection_name: str,
    normal_data: dict[str, list[Formula]],
    grading_data: dict[str, list[Formula]],
    reportable_grades: list[int] | None = None,
    reportable_grades_exceptions: dict[str, list[int]] | None = None,
    keep_existing: bool | None = None,
    create_missing_normal: bool | None = None,
) -> ReferenceRangeCollection:
    """Load the reference ranges for a single collection.

    See also: load_all_reference_ranges
    """
    (
        reference_range_collection,
        _,
    ) = reference_range_colllection_model_cls().objects.get_or_create(name=collection_name)
    reportable_grades = reportable_grades or get_default_reportable_grades()
    for grade in reportable_grades:
        setattr(reference_range_collection, f"grade{grade}", True)
    reference_range_collection.save()

    update_grading_exceptions(
        reference_range_collection=reference_range_collection,
        reportable_grades_exceptions=reportable_grades_exceptions,
        keep_existing=keep_existing,
    )

    update_normal_data(reference_range_collection, normal_data=normal_data)

    update_grading_data(
        reference_range_collection,
        grading_data=grading_data,
        reportable_grades=reportable_grades,
        reportable_grades_exceptions=reportable_grades_exceptions,
        create_missing_normal=True,
    )
    return reference_range_collection



from django.contrib import admin
from django.template.loader import render_to_string
from django.urls import reverse
from django_audit_fields.admin import audit_fieldset_tuple
from rangefilter.filters import DateRangeFilterBuilder

from edc_model_admin.history import SimpleHistoryAdmin
from edc_sites.admin import SiteModelAdminMixin
from edc_utils.date import to_local

from ...admin_site import edc_pharmacy_admin
from ...forms import StorageBinItemForm
from ...models import StorageBinItem
from ..model_admin_mixin import ModelAdminMixin


@admin.register(StorageBinItem, site=edc_pharmacy_admin)
class StorageBinItemAdmin(SiteModelAdminMixin, ModelAdminMixin, SimpleHistoryAdmin):
    change_list_title = "Pharmacy: Storage bin item"
    change_form_title = "Pharmacy: Storage bin item"
    history_list_display = ()
    show_object_tools = True
    show_cancel = True
    list_per_page = 20

    autocomplete_fields = ["stock"]
    actions = ["delete_selected"]

    change_list_note = "Once an item is dispensed it is automatically removed from storage."

    form = StorageBinItemForm

    fieldsets = (
        (
            None,
            {
                "fields": (
                    "storage_bin",
                    "stock",
                    "item_datetime",
                )
            },
        ),
        audit_fieldset_tuple,
    )

    list_display = (
        "item_identifier",
        "storage_bin_changelist",
        "stock_changelist",
        "subject",
        "location",
    )

    search_fields = (
        "id",
        "item_identifier",
        "stock__allocation__registered_subject__subject_identifier",
        "stock__code",
        "storage_bin__bin_identifier",
        "storage_bin__name",
        "storage_bin__id",
    )

    list_filter = (
        "storage_bin__in_use",
        "storage_bin__location",
        ("item_datetime", DateRangeFilterBuilder()),
    )

    @admin.display(description="Item #", ordering="-item_identifier")
    def identifier(self, obj):
        return obj.item_identifier

    @admin.display(description="Location", ordering="storage_bin__location__name")
    def location(self, obj):
        return obj.storage_bin.location

    @admin.display(description="Item date", ordering="item_datetime")
    def item_date(self, obj):
        return to_local(obj.item_datetime).date()

    @admin.display(description="Bin", ordering="storage_bin__name")
    def storage_bin_changelist(self, obj):
        bin_ref = obj.storage_bin.name or obj.storage_bin.bin_identifier
        url = reverse("edc_pharmacy_admin:edc_pharmacy_storagebin_changelist")
        url = f"{url}?q={bin_ref}"
        context = dict(url=url, label=bin_ref, title="Go to bin")
        return render_to_string("edc_pharmacy/stock/items_as_link.html", context=context)

    @admin.display(description="Stock")
    def stock_changelist(self, obj):
        url = reverse("edc_pharmacy_admin:edc_pharmacy_stock_changelist")
        url = f"{url}?q={obj.stock.code}"
        context = dict(url=url, label=obj.stock.code, title="Go to stock")
        return render_to_string("edc_pharmacy/stock/items_as_link.html", context=context)

    @admin.display(
        description="Subject #",
        ordering="stock__allocation__registered_subject__subject_identifier",
    )
    def subject(self, obj):
        return obj.stock.allocation.registered_subject.subject_identifier

    def get_queryset(self, request):
        qs = super().get_queryset(request)
        return qs.filter(stock__dispenseitem__isnull=True)

    def get_view_only_site_ids_for_user(self, request) -> list[int]:
        return [s.id for s in request.user.userprofile.sites.all() if s.id != request.site.id]



from dataclasses import dataclass, field


@dataclass
class Address:
    contact_name: str = field(default="CONTACT NAME")
    company_name: str = field(default="COMPANY NAME")
    address: str = field(default="ADDRESS")
    city: str = field(default="CITY")
    state: str = field(default="")
    postal_code: str = field(default="0000")
    country: str = field(default="COUNTRY")
    tel: str = field(default="TELEPHONE")
    mobile: str = field(default="MOBILE")
    fax: str = field(default="FAX")



from __future__ import annotations

from dataclasses import dataclass, field
from typing import TYPE_CHECKING, TypeVar
from uuid import UUID

from django.contrib.sites.models import Site
from django.db import models
from django.utils.translation import gettext as _

from edc_view_utils import DashboardModelButton

if TYPE_CHECKING:
    from edc_screening.model_mixins import ScreeningModelMixin

    ScreeningModel = TypeVar("ScreeningModel", bound=ScreeningModelMixin)

__all__ = ["SubjectScreeningButton"]


@dataclass
class SubjectScreeningButton(DashboardModelButton):
    model_obj: ScreeningModel = None
    metadata_model_obj: models.Model = field(init=False)

    def __post_init__(self):
        self.model_cls = self.model_obj.__class__

    @property
    def color(self) -> str:
        return "success"

    @property
    def fa_icon(self) -> str:
        if self.perms.view_only:
            fa_icon = "fa-eye"
        else:
            fa_icon = "fa-eye" if self.model_obj.consented else "fa-pencil"
        return fa_icon

    @property
    def label(self) -> str:
        return ""

    @property
    def site(self) -> Site:
        return self.model_obj.site

    @property
    def reverse_kwargs(self) -> dict[str, str | UUID]:
        return dict(screening_identifier=self.model_obj.screening_identifier)

    @property
    def title(self) -> str:
        title = _("View") if self.perms.view_only or self.model_obj.consented else _("Edit")
        verbose_name = self.model_cls._meta.verbose_name.lower()
        return f"{title} {verbose_name}"



from django.contrib.sites.shortcuts import get_current_site
from django.core.handlers.wsgi import WSGIRequest

from ..exceptions import InvalidSiteError


def is_current_site_or_raise(site_id: int, request: WSGIRequest = None) -> bool:
    if site_id != get_current_site(request).id:
        raise InvalidSiteError(
            f"Expected the current site. Current site is {get_current_site(request).id}. "
            f"Got {site_id}."
        )
    return True



import sys

from .table import Table


class Consent(Table):
    default_columns = [
        "subject_identifier",
        "gender",
        "dob",
        "consent_datetime",
        "study_site",
    ]

    def prepare(self, **kwargs):
        self.dataframe["dob"] = self.helper.date_to_local_datetime(self.dataframe["dob"])
        self.dataframe["consent_datetime"] = self.helper.to_local_datetime(
            self.dataframe["consent_datetime"]
        )
        self.dataframe = self.dataframe.sort_values(["subject_identifier"])
        self.dataframe = self.dataframe.drop_duplicates()
        self.dataframe = (
            self.dataframe.groupby(["subject_identifier", "consent_datetime"])
            .min()
            .reset_index()
        )
        self.dataframe["duplicate"] = self.dataframe.duplicated(self.dataframe)
        if len(self.dataframe[self.dataframe["duplicate"]]) > 0:
            self.dataframe = self.deduplicate()
        self.dataframe["consent_date"] = self.dataframe["consent_datetime"].dt.normalize()
        self.columns.append("consent_date")

    def deduplicate(self):
        sys.stderr.write("Consent has duplicates!")
        sys.stderr.write(self.dataframe[self.dataframe["duplicate"]])

    def read_csv(self, **kwargs):
        super().read_csv(**kwargs)
        self.dataframe["consent_datetime"] = self.dataframe["consent_datetime"].astype(
            "datetime64[ns]"
        )
        self.dataframe["consent_date"] = self.dataframe["consent_date"].astype(
            "datetime64[ns]"
        )



# Generated by Django 4.2.3 on 2023-07-25 18:31

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("edc_facility", "0005_healthfacility_healthfacilitytypes_and_more"),
    ]

    operations = [
        migrations.AlterField(
            model_name="healthfacility",
            name="health_facility_type",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.PROTECT,
                related_name="+",
                to="edc_facility.healthfacilitytypes",
                verbose_name="Health facility type",
            ),
        ),
    ]



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from ..create_refill import create_refill

if TYPE_CHECKING:
    from ...models import RxRefill


def calculate_days_to_next_refill(refill) -> int:
    """Returns the number of days until medication runs out"""
    return 0


def create_refills_from_crf(instance: Any, related_visit_model_attr: str) -> RxRefill:
    subject_visit = getattr(instance, related_visit_model_attr)
    rx_refill = create_refill(
        refill_identifier=instance.refill_identifier,
        subject_identifier=subject_visit.subject_identifier,
        dosage_guideline=instance.dosage_guideline,
        formulation=instance.formulation,
        refill_start_datetime=instance.refill_start_datetime,
        refill_end_datetime=instance.refill_end_datetime,
        roundup_divisible_by=instance.roundup_divisible_by,
        weight_in_kgs=getattr(instance, "weight_in_kgs", None),
    )
    return rx_refill



# Generated by Django 2.0rc1 on 2017-11-27 15:41

import _socket
import django.db.models.deletion
import django.utils.timezone
import django_audit_fields.fields.uuid_auto_field
import django_revision.revision_field
from django.conf import settings
from django.db import migrations, models

import edc_model.models
import edc_model_fields.fields.hostname_modification_field
import edc_model_fields.fields.userfield
import django.utils.timezone


class Migration(migrations.Migration):
    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ("edc_lab", "0009_auto_20171116_1606"),
    ]

    operations = [
        migrations.CreateModel(
            name="HistoricalOrder",
            fields=[
                (
                    "created",
                    models.DateTimeField(blank=True, default=django.utils.timezone.now),
                ),
                (
                    "modified",
                    models.DateTimeField(blank=True, default=django.utils.timezone.now),
                ),
                (
                    "user_created",
                    edc_model_fields.fields.userfield.UserField(
                        blank=True,
                        help_text="Updated by admin.save_model",
                        max_length=50,
                        verbose_name="user created",
                    ),
                ),
                (
                    "user_modified",
                    edc_model_fields.fields.userfield.UserField(
                        blank=True,
                        help_text="Updated by admin.save_model",
                        max_length=50,
                        verbose_name="user modified",
                    ),
                ),
                (
                    "hostname_created",
                    models.CharField(
                        blank=True,
                        default=_socket.gethostname,
                        help_text="System field. (modified on create only)",
                        max_length=60,
                    ),
                ),
                (
                    "hostname_modified",
                    edc_model_fields.fields.hostname_modification_field.HostnameModificationField(
                        blank=True,
                        help_text="System field. (modified on every save)",
                        max_length=50,
                    ),
                ),
                (
                    "revision",
                    django_revision.revision_field.RevisionField(
                        blank=True,
                        editable=False,
                        help_text="System field. Git repository tag:branch:commit.",
                        max_length=75,
                        null=True,
                        verbose_name="Revision",
                    ),
                ),
                ("device_created", models.CharField(blank=True, max_length=10)),
                ("device_modified", models.CharField(blank=True, max_length=10)),
                (
                    "id",
                    django_audit_fields.fields.uuid_auto_field.UUIDAutoField(
                        blank=True,
                        db_index=True,
                        editable=False,
                        help_text="System auto field. UUID primary key.",
                    ),
                ),
                (
                    "order_identifier",
                    models.CharField(db_index=True, editable=False, max_length=25),
                ),
                (
                    "order_datetime",
                    models.DateTimeField(
                        default=django.utils.timezone.now,
                        validators=[edc_model.validators.datetime_not_future],
                    ),
                ),
                ("panel_name", models.CharField(max_length=25)),
                ("history_date", models.DateTimeField()),
                ("history_change_reason", models.CharField(max_length=100, null=True)),
                (
                    "history_type",
                    models.CharField(
                        choices=[("+", "Created"), ("~", "Changed"), ("-", "Deleted")],
                        max_length=1,
                    ),
                ),
                (
                    "history_id",
                    django_audit_fields.fields.uuid_auto_field.UUIDAutoField(
                        primary_key=True, serialize=False
                    ),
                ),
                (
                    "aliquot",
                    models.ForeignKey(
                        blank=True,
                        db_constraint=False,
                        null=True,
                        on_delete=django.db.models.deletion.DO_NOTHING,
                        related_name="+",
                        to="edc_lab.Aliquot",
                    ),
                ),
                (
                    "history_user",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name="+",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
            options={
                "verbose_name": "historical ",
                "ordering": ("-history_date", "-history_id"),
                "get_latest_by": "history_date",
            },
        ),
        migrations.CreateModel(
            name="HistoricalResult",
            fields=[
                (
                    "created",
                    models.DateTimeField(blank=True, default=django.utils.timezone.now),
                ),
                (
                    "modified",
                    models.DateTimeField(blank=True, default=django.utils.timezone.now),
                ),
                (
                    "user_created",
                    edc_model_fields.fields.userfield.UserField(
                        blank=True,
                        help_text="Updated by admin.save_model",
                        max_length=50,
                        verbose_name="user created",
                    ),
                ),
                (
                    "user_modified",
                    edc_model_fields.fields.userfield.UserField(
                        blank=True,
                        help_text="Updated by admin.save_model",
                        max_length=50,
                        verbose_name="user modified",
                    ),
                ),
                (
                    "hostname_created",
                    models.CharField(
                        blank=True,
                        default=_socket.gethostname,
                        help_text="System field. (modified on create only)",
                        max_length=60,
                    ),
                ),
                (
                    "hostname_modified",
                    edc_model_fields.fields.hostname_modification_field.HostnameModificationField(
                        blank=True,
                        help_text="System field. (modified on every save)",
                        max_length=50,
                    ),
                ),
                (
                    "revision",
                    django_revision.revision_field.RevisionField(
                        blank=True,
                        editable=False,
                        help_text="System field. Git repository tag:branch:commit.",
                        max_length=75,
                        null=True,
                        verbose_name="Revision",
                    ),
                ),
                ("device_created", models.CharField(blank=True, max_length=10)),
                ("device_modified", models.CharField(blank=True, max_length=10)),
                (
                    "id",
                    django_audit_fields.fields.uuid_auto_field.UUIDAutoField(
                        blank=True,
                        db_index=True,
                        editable=False,
                        help_text="System auto field. UUID primary key.",
                    ),
                ),
                ("panel_name", models.CharField(max_length=25)),
                ("report_datetime", models.DateTimeField(null=True)),
                (
                    "pending_datetime",
                    models.DateTimeField(default=django.utils.timezone.now),
                ),
                ("pending", models.BooleanField(default=True)),
                (
                    "resulted_datetime",
                    models.DateTimeField(default=django.utils.timezone.now),
                ),
                ("resulted", models.BooleanField(default=False)),
                ("history_date", models.DateTimeField()),
                ("history_change_reason", models.CharField(max_length=100, null=True)),
                (
                    "history_type",
                    models.CharField(
                        choices=[("+", "Created"), ("~", "Changed"), ("-", "Deleted")],
                        max_length=1,
                    ),
                ),
                (
                    "history_id",
                    django_audit_fields.fields.uuid_auto_field.UUIDAutoField(
                        primary_key=True, serialize=False
                    ),
                ),
                (
                    "history_user",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name="+",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
            options={
                "verbose_name": "historical ",
                "ordering": ("-history_date", "-history_id"),
                "get_latest_by": "history_date",
            },
        ),
        migrations.CreateModel(
            name="HistoricalResultItem",
            fields=[
                (
                    "created",
                    models.DateTimeField(blank=True, default=django.utils.timezone.now),
                ),
                (
                    "modified",
                    models.DateTimeField(blank=True, default=django.utils.timezone.now),
                ),
                (
                    "user_created",
                    edc_model_fields.fields.userfield.UserField(
                        blank=True,
                        help_text="Updated by admin.save_model",
                        max_length=50,
                        verbose_name="user created",
                    ),
                ),
                (
                    "user_modified",
                    edc_model_fields.fields.userfield.UserField(
                        blank=True,
                        help_text="Updated by admin.save_model",
                        max_length=50,
                        verbose_name="user modified",
                    ),
                ),
                (
                    "hostname_created",
                    models.CharField(
                        blank=True,
                        default=_socket.gethostname,
                        help_text="System field. (modified on create only)",
                        max_length=60,
                    ),
                ),
                (
                    "hostname_modified",
                    edc_model_fields.fields.hostname_modification_field.HostnameModificationField(
                        blank=True,
                        help_text="System field. (modified on every save)",
                        max_length=50,
                    ),
                ),
                (
                    "revision",
                    django_revision.revision_field.RevisionField(
                        blank=True,
                        editable=False,
                        help_text="System field. Git repository tag:branch:commit.",
                        max_length=75,
                        null=True,
                        verbose_name="Revision",
                    ),
                ),
                ("device_created", models.CharField(blank=True, max_length=10)),
                ("device_modified", models.CharField(blank=True, max_length=10)),
                (
                    "id",
                    django_audit_fields.fields.uuid_auto_field.UUIDAutoField(
                        blank=True,
                        db_index=True,
                        editable=False,
                        help_text="System auto field. UUID primary key.",
                    ),
                ),
                ("report_datetime", models.DateTimeField(null=True)),
                ("utestid", models.CharField(max_length=25, null=True)),
                ("value", models.CharField(max_length=25, null=True)),
                ("quantifier", models.CharField(max_length=25, null=True)),
                ("value_datetime", models.DateTimeField(null=True)),
                ("reference", models.CharField(max_length=25, null=True)),
                (
                    "pending_datetime",
                    models.DateTimeField(default=django.utils.timezone.now),
                ),
                ("pending", models.BooleanField(default=True)),
                ("resulted_datetime", models.DateTimeField(null=True)),
                ("resulted", models.BooleanField(default=False)),
                ("history_date", models.DateTimeField()),
                ("history_change_reason", models.CharField(max_length=100, null=True)),
                (
                    "history_type",
                    models.CharField(
                        choices=[("+", "Created"), ("~", "Changed"), ("-", "Deleted")],
                        max_length=1,
                    ),
                ),
                (
                    "history_id",
                    django_audit_fields.fields.uuid_auto_field.UUIDAutoField(
                        primary_key=True, serialize=False
                    ),
                ),
                (
                    "history_user",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name="+",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
            options={
                "verbose_name": "historical ",
                "ordering": ("-history_date", "-history_id"),
                "get_latest_by": "history_date",
            },
        ),
        migrations.CreateModel(
            name="Order",
            fields=[
                (
                    "created",
                    models.DateTimeField(blank=True, default=django.utils.timezone.now),
                ),
                (
                    "modified",
                    models.DateTimeField(blank=True, default=django.utils.timezone.now),
                ),
                (
                    "user_created",
                    edc_model_fields.fields.userfield.UserField(
                        blank=True,
                        help_text="Updated by admin.save_model",
                        max_length=50,
                        verbose_name="user created",
                    ),
                ),
                (
                    "user_modified",
                    edc_model_fields.fields.userfield.UserField(
                        blank=True,
                        help_text="Updated by admin.save_model",
                        max_length=50,
                        verbose_name="user modified",
                    ),
                ),
                (
                    "hostname_created",
                    models.CharField(
                        blank=True,
                        default=_socket.gethostname,
                        help_text="System field. (modified on create only)",
                        max_length=60,
                    ),
                ),
                (
                    "hostname_modified",
                    edc_model_fields.fields.hostname_modification_field.HostnameModificationField(
                        blank=True,
                        help_text="System field. (modified on every save)",
                        max_length=50,
                    ),
                ),
                (
                    "revision",
                    django_revision.revision_field.RevisionField(
                        blank=True,
                        editable=False,
                        help_text="System field. Git repository tag:branch:commit.",
                        max_length=75,
                        null=True,
                        verbose_name="Revision",
                    ),
                ),
                ("device_created", models.CharField(blank=True, max_length=10)),
                ("device_modified", models.CharField(blank=True, max_length=10)),
                (
                    "id",
                    django_audit_fields.fields.uuid_auto_field.UUIDAutoField(
                        blank=True,
                        editable=False,
                        help_text="System auto field. UUID primary key.",
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                (
                    "order_identifier",
                    models.CharField(editable=False, max_length=25, unique=True),
                ),
                (
                    "order_datetime",
                    models.DateTimeField(
                        default=django.utils.timezone.now,
                        validators=[edc_model.validators.datetime_not_future],
                    ),
                ),
                ("panel_name", models.CharField(max_length=25)),
                (
                    "aliquot",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.PROTECT,
                        to="edc_lab.Aliquot",
                    ),
                ),
            ],
            options={
                "ordering": ("-modified", "-created"),
                "get_latest_by": "modified",
                "abstract": False,
            },
        ),
        migrations.CreateModel(
            name="Result",
            fields=[
                (
                    "created",
                    models.DateTimeField(blank=True, default=django.utils.timezone.now),
                ),
                (
                    "modified",
                    models.DateTimeField(blank=True, default=django.utils.timezone.now),
                ),
                (
                    "user_created",
                    edc_model_fields.fields.userfield.UserField(
                        blank=True,
                        help_text="Updated by admin.save_model",
                        max_length=50,
                        verbose_name="user created",
                    ),
                ),
                (
                    "user_modified",
                    edc_model_fields.fields.userfield.UserField(
                        blank=True,
                        help_text="Updated by admin.save_model",
                        max_length=50,
                        verbose_name="user modified",
                    ),
                ),
                (
                    "hostname_created",
                    models.CharField(
                        blank=True,
                        default=_socket.gethostname,
                        help_text="System field. (modified on create only)",
                        max_length=60,
                    ),
                ),
                (
                    "hostname_modified",
                    edc_model_fields.fields.hostname_modification_field.HostnameModificationField(
                        blank=True,
                        help_text="System field. (modified on every save)",
                        max_length=50,
                    ),
                ),
                (
                    "revision",
                    django_revision.revision_field.RevisionField(
                        blank=True,
                        editable=False,
                        help_text="System field. Git repository tag:branch:commit.",
                        max_length=75,
                        null=True,
                        verbose_name="Revision",
                    ),
                ),
                ("device_created", models.CharField(blank=True, max_length=10)),
                ("device_modified", models.CharField(blank=True, max_length=10)),
                (
                    "id",
                    django_audit_fields.fields.uuid_auto_field.UUIDAutoField(
                        blank=True,
                        editable=False,
                        help_text="System auto field. UUID primary key.",
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                ("panel_name", models.CharField(max_length=25)),
                ("report_datetime", models.DateTimeField(null=True)),
                (
                    "pending_datetime",
                    models.DateTimeField(default=django.utils.timezone.now),
                ),
                ("pending", models.BooleanField(default=True)),
                (
                    "resulted_datetime",
                    models.DateTimeField(default=django.utils.timezone.now),
                ),
                ("resulted", models.BooleanField(default=False)),
                (
                    "order",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.PROTECT, to="edc_lab.Order"
                    ),
                ),
            ],
            options={"abstract": False},
        ),
        migrations.CreateModel(
            name="ResultItem",
            fields=[
                (
                    "created",
                    models.DateTimeField(blank=True, default=django.utils.timezone.now),
                ),
                (
                    "modified",
                    models.DateTimeField(blank=True, default=django.utils.timezone.now),
                ),
                (
                    "user_created",
                    edc_model_fields.fields.userfield.UserField(
                        blank=True,
                        help_text="Updated by admin.save_model",
                        max_length=50,
                        verbose_name="user created",
                    ),
                ),
                (
                    "user_modified",
                    edc_model_fields.fields.userfield.UserField(
                        blank=True,
                        help_text="Updated by admin.save_model",
                        max_length=50,
                        verbose_name="user modified",
                    ),
                ),
                (
                    "hostname_created",
                    models.CharField(
                        blank=True,
                        default=_socket.gethostname,
                        help_text="System field. (modified on create only)",
                        max_length=60,
                    ),
                ),
                (
                    "hostname_modified",
                    edc_model_fields.fields.hostname_modification_field.HostnameModificationField(
                        blank=True,
                        help_text="System field. (modified on every save)",
                        max_length=50,
                    ),
                ),
                (
                    "revision",
                    django_revision.revision_field.RevisionField(
                        blank=True,
                        editable=False,
                        help_text="System field. Git repository tag:branch:commit.",
                        max_length=75,
                        null=True,
                        verbose_name="Revision",
                    ),
                ),
                ("device_created", models.CharField(blank=True, max_length=10)),
                ("device_modified", models.CharField(blank=True, max_length=10)),
                (
                    "id",
                    django_audit_fields.fields.uuid_auto_field.UUIDAutoField(
                        blank=True,
                        editable=False,
                        help_text="System auto field. UUID primary key.",
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                ("report_datetime", models.DateTimeField(null=True)),
                ("utestid", models.CharField(max_length=25, null=True)),
                ("value", models.CharField(max_length=25, null=True)),
                ("quantifier", models.CharField(max_length=25, null=True)),
                ("value_datetime", models.DateTimeField(null=True)),
                ("reference", models.CharField(max_length=25, null=True)),
                (
                    "pending_datetime",
                    models.DateTimeField(default=django.utils.timezone.now),
                ),
                ("pending", models.BooleanField(default=True)),
                ("resulted_datetime", models.DateTimeField(null=True)),
                ("resulted", models.BooleanField(default=False)),
                (
                    "result",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.PROTECT, to="edc_lab.Result"
                    ),
                ),
            ],
            options={"abstract": False},
        ),
        migrations.AddField(
            model_name="historicalresultitem",
            name="result",
            field=models.ForeignKey(
                blank=True,
                db_constraint=False,
                null=True,
                on_delete=django.db.models.deletion.DO_NOTHING,
                related_name="+",
                to="edc_lab.Result",
            ),
        ),
        migrations.AddField(
            model_name="historicalresult",
            name="order",
            field=models.ForeignKey(
                blank=True,
                db_constraint=False,
                null=True,
                on_delete=django.db.models.deletion.DO_NOTHING,
                related_name="+",
                to="edc_lab.Order",
            ),
        ),
    ]



from django.apps import apps as django_apps

from edc_auth.constants import (
    ADMINISTRATION,
    AUDITOR_ROLE,
    CLINICIAN_ROLE,
    EVERYONE,
    NURSE_ROLE,
    PII_VIEW,
)
from edc_auth.site_auths import site_auths
from edc_export.constants import EXPORT

from .auth_objects import (
    LAB,
    LAB_TECHNICIAN_ROLE,
    LAB_VIEW,
    lab_codenames,
    lab_view_codenames,
)


def update_site_auths():
    site_auths.add_group(*lab_codenames, name=LAB)
    site_auths.add_group(*lab_view_codenames, name=LAB_VIEW)

    if django_apps.is_installed("edc_export"):
        site_auths.update_group(
            "edc_lab.export_aliquot",
            "edc_lab.export_box",
            "edc_lab.export_boxitem",
            "edc_lab.export_boxtype",
            "edc_lab.export_order",
            "edc_lab.export_panel",
            "edc_lab.export_result",
            "edc_lab.export_resultitem",
            name=EXPORT,
        )

    site_auths.add_role(ADMINISTRATION, EVERYONE, LAB, PII_VIEW, name=LAB_TECHNICIAN_ROLE)
    site_auths.update_role(LAB, name=CLINICIAN_ROLE)
    site_auths.update_role(LAB, name=NURSE_ROLE)
    site_auths.update_role(LAB_VIEW, name=AUDITOR_ROLE)


update_site_auths()



from .get_appointments import get_appointments


def get_next_appointments(**kwargs):
    """Returns a dataframe for the next appointment in the
    appointment schedule (first record where appt_status=NEW_APPT).

    To show appts where next appt is a future date:
        df = get_next_appointments()
        df[df.next_appt_datetime>=pd.to_datetime("today")]

    """
    df = get_appointments(**kwargs)
    df = (
        df.groupby(
            by=[
                "subject_identifier",
                "site_id",
                "next_visit_code",
                "next_appt_datetime",
            ]
        )
        .size()
        .to_frame()
        .reset_index()
    )
    df = df[["subject_identifier", "site_id", "next_visit_code", "next_appt_datetime"]]
    return df



from .allocation import Allocation
from .allocation_proxy import AllocationProxy
from .confirmation import Confirmation
from .confirmation_at_site import ConfirmationAtSite
from .confirmation_at_site_item import ConfirmationAtSiteItem
from .container import Container
from .container_type import ContainerType
from .container_units import ContainerUnits
from .dispense import Dispense
from .dispense_item import DispenseItem
from .location import Location
from .lot import Lot
from .order import Order
from .order_item import OrderItem
from .product import Product
from .receive import Receive
from .receive_item import ReceiveItem
from .repack_request import RepackRequest
from .stock import Stock
from .stock_adjustment import StockAdjustment
from .stock_proxy import StockProxy
from .stock_request import StockRequest
from .stock_request_item import StockRequestItem
from .stock_transfer import StockTransfer
from .stock_transfer_item import StockTransferItem
from .storage_bin import StorageBin
from .storage_bin_item import StorageBinItem
from .supplier import Supplier



from django.db import models
from django.utils import timezone

from edc_constants.choices import YES_NO
from edc_constants.constants import NOT_APPLICABLE, YES
from edc_model.validators import date_not_future

from ...choices import AE_GRADE_SIMPLE, AE_OUTCOME
from ...utils import get_adverse_event_app_label


class AeFollowupFieldsModelMixin(models.Model):
    ae_initial = models.ForeignKey(
        f"{get_adverse_event_app_label()}.aeinitial", on_delete=models.PROTECT
    )

    report_datetime = models.DateTimeField(
        verbose_name="Report date and time", default=timezone.now
    )

    outcome = models.CharField(blank=False, null=False, max_length=25, choices=AE_OUTCOME)

    outcome_date = models.DateField(validators=[date_not_future])

    ae_grade = models.CharField(
        verbose_name="If severity increased, indicate grade",
        max_length=25,
        choices=AE_GRADE_SIMPLE,
        default=NOT_APPLICABLE,
    )

    relevant_history = models.TextField(
        verbose_name="Description summary of Adverse Event outcome",
        max_length=1000,
        blank=False,
        null=False,
        help_text="Indicate Adverse Event, clinical results,"
        "medications given, dosage,treatment plan and outcomes.",
    )

    followup = models.CharField(
        verbose_name="Is a follow-up to this report required?",
        max_length=15,
        choices=YES_NO,
        default=YES,
        help_text="If NO, this will be considered the final report",
    )

    class Meta:
        abstract = True



from .age_evaluator import AgeEvaluator
from .gender_evaluator import GenderEvaluator


class EligibilityError(Exception):
    pass


class Eligibility:
    """Eligible if all criteria evaluate True.

    Any key in `additional_criteria` has value True if eligible.
    """

    # default to M or F
    gender_evaluator_cls = GenderEvaluator

    # default to eligible if >=18
    age_evaluator = AgeEvaluator(age_lower=18, age_lower_inclusive=True)

    custom_reasons_dict: dict = {}

    def __init__(
        self,
        age: int = None,
        gender: str = None,
        pregnant: bool = None,
        breast_feeding: bool = None,
        **additional_criteria,
    ) -> None:
        self.criteria = dict(**additional_criteria)
        if len(self.criteria) == 0:
            raise EligibilityError("No criteria provided.")

        self.gender_evaluator = self.gender_evaluator_cls(
            gender=gender, pregnant=pregnant, breast_feeding=breast_feeding
        )
        self.criteria.update(age=self.age_evaluator.eligible(age))
        self.criteria.update(gender=self.gender_evaluator.eligible)

        # hook for custom checks
        self.criteria.update(**self.extra_eligibility_criteria)

        # eligible if all criteria are True
        self.eligible = all([v for v in self.criteria.values()])

        if self.eligible:
            self.reasons_ineligible = ""
        else:
            self.reasons_ineligible = {k: v for k, v in self.criteria.items() if not v}
            for k, v in self.criteria.items():
                if not v:
                    if k in self.get_custom_reasons_dict():
                        self.reasons_ineligible.update(
                            {k: self.get_custom_reasons_dict().get(k)}
                        )
                    elif k not in ["age", "gender"]:
                        self.reasons_ineligible.update({k: k})
            if not self.age_evaluator.eligible(age):
                self.reasons_ineligible.update(age=self.age_evaluator.reasons_ineligible)
            if not self.gender_evaluator.eligible:
                self.reasons_ineligible.update(
                    gender=f"{' and '.join(self.gender_evaluator.reasons_ineligible)}."
                )

    def __str__(self):
        return self.eligible

    @property
    def extra_eligibility_criteria(self) -> dict:
        return {}

    def get_custom_reasons_dict(self) -> dict:
        """Returns a dictionary of custom reasons for named criteria."""
        for k in self.custom_reasons_dict:
            if k in self.custom_reasons_dict and k not in self.criteria:
                raise EligibilityError(
                    f"Custom reasons refer to invalid named criteria, Got '{k}'. "
                    f"Expected one of {list(self.criteria)}. "
                    f"See {self!r}."
                )
        return self.custom_reasons_dict



# Generated by Django 5.1.2 on 2024-11-21 03:43

from django.db import migrations, models

import django.utils.timezone


class Migration(migrations.Migration):
    dependencies = [
        ("edc_pharmacy", "0050_remove_stocktransferconfirmation2_location_and_more"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="historicalstocktransferconfirmationitem",
            options={
                "get_latest_by": ("history_date", "history_id"),
                "ordering": ("-history_date", "-history_id"),
                "verbose_name": "historical Stock Transfer Confirmation Item",
                "verbose_name_plural": "historical Stock Transfer Confirmation Items",
            },
        ),
        migrations.AlterModelOptions(
            name="stocktransferconfirmationitem",
            options={
                "default_manager_name": "objects",
                "default_permissions": (
                    "add",
                    "change",
                    "delete",
                    "view",
                    "export",
                    "import",
                ),
                "verbose_name": "Stock Transfer Confirmation Item",
                "verbose_name_plural": "Stock Transfer Confirmation Items",
            },
        ),
        migrations.RenameField(
            model_name="historicalstocktransferconfirmation",
            old_name="confirmation_identifier",
            new_name="transfer_confirmation_identifier",
        ),
        migrations.RenameField(
            model_name="historicalstocktransferconfirmationitem",
            old_name="confirmation_item_datetime",
            new_name="transfer_confirmation_item_datetime",
        ),
        migrations.RenameField(
            model_name="historicalstocktransferconfirmationitem",
            old_name="confirmation_item_identifier",
            new_name="transfer_confirmation_item_identifier",
        ),
        migrations.RenameField(
            model_name="stocktransferconfirmation",
            old_name="confirmation_identifier",
            new_name="transfer_confirmation_identifier",
        ),
        migrations.RenameField(
            model_name="stocktransferconfirmationitem",
            old_name="confirmation_item_datetime",
            new_name="transfer_confirmation_item_datetime",
        ),
        migrations.RenameField(
            model_name="stocktransferconfirmationitem",
            old_name="confirmation_item_identifier",
            new_name="transfer_confirmation_item_identifier",
        ),
        migrations.RemoveField(
            model_name="historicalstocktransferconfirmation",
            name="confirmation_datetime",
        ),
        migrations.RemoveField(
            model_name="stocktransferconfirmation",
            name="confirmation_datetime",
        ),
        migrations.AddField(
            model_name="historicalstocktransferconfirmation",
            name="transfer_confirmation_datetime",
            field=models.DateTimeField(default=django.utils.timezone.now),
        ),
        migrations.AddField(
            model_name="stocktransferconfirmation",
            name="transfer_confirmation_datetime",
            field=models.DateTimeField(default=django.utils.timezone.now),
        ),
    ]



from .blood_result_model_mixin import (
    BloodResultsFieldsModelMixin,
    BloodResultsMethodsModelMixin,
    BloodResultsModelMixin,
)
from .electrolytes_model_mixins import (
    MagnesiumModelMixin,
    PotassiumModelMixin,
    SodiumModelMixin,
)
from .fbc_model_mixins import (
    HaemoglobinModelMixin,
    HctModelMixin,
    LymphocyteDiffModelMixin,
    LymphocyteModelMixin,
    MchcModelMixin,
    MchModelMixin,
    McvModelMixin,
    NeutrophilDiffModelMixin,
    NeutrophilModelMixin,
    PlateletsModelMixin,
    RbcModelMixin,
    WbcModelMixin,
)
from .fbg_model_mixin import FbgModelMixin
from .glucose_model_mixin import GlucoseModelMixin
from .hba1c_model_mixin import Hba1cModelMixin
from .insulin_model_mixin import InsulinModelMixin
from .lft_model_mixins import (
    AlbuminModelMixin,
    AlpModelMixin,
    AltModelMixin,
    AmylaseModelMixin,
    AstModelMixin,
    CrpModelMixin,
    GgtModelMixin,
    TotalBilirubinModelMixin,
)
from .lipid_model_mixins import (
    CholModelMixin,
    HdlModelMixin,
    LdlModelMixin,
    TrigModelMixin,
)
from .proteinuria_model_mixin import ProteinuriaModelMixin
from .rft_model_mixins import CreatinineModelMixin, UreaModelMixin, UricAcidModelMixin



from django.contrib import messages
from django.core.exceptions import ObjectDoesNotExist
from django.urls import reverse
from django.utils.html import format_html

from edc_dashboard.url_names import url_names
from edc_dashboard.view_mixins import EdcViewMixin
from edc_lab import SHIPPED
from edc_lab.models import BoxIsFullError, BoxItem

from ...view_mixins import BoxViewMixin
from .action_view import ActionView


class ManageBoxItemView(EdcViewMixin, BoxViewMixin, ActionView):
    post_action_url = "manage_box_listboard_url"
    valid_form_actions = ["add_item", "renumber_items", "remove_selected_items"]

    @property
    def url_kwargs(self):
        return {
            "action_name": self.kwargs.get("action_name"),
            "box_identifier": self.box_identifier,
        }

    def process_form_action(self, request=None):
        if self.action == "add_item":
            if self.box_item_identifier:
                self.add_box_item()
        elif self.action == "renumber_items":
            self.renumber_items()
        elif self.action == "remove_selected_items":
            self.remove_selected_items()

    def remove_selected_items(self):
        """Deletes the selected items."""
        if not self.selected_items:
            message = "Nothing to do. No items have been selected."
            messages.warning(self.request, message)
        elif self.box.status == SHIPPED:
            message = "Unable to remove. Box has already been shipped."
            messages.error(self.request, message)
        else:
            deleted = (
                BoxItem.objects.filter(pk__in=self.selected_items)
                .exclude(box__status=SHIPPED)
                .delete()
            )
            message = f"{deleted[0]} items have been removed."
            messages.success(self.request, message)

    def renumber_items(self):
        """Resets positions to be a sequence incremented by 1."""
        box_items = self.box.boxitem_set.all().order_by("position")
        if box_items.count() == 0:
            message = "Nothing to do. There are no items in the box."
            messages.warning(self.request, message)
        elif self.box.status == SHIPPED:
            message = "Unable to renumber. Box has already been shipped."
            messages.error(self.request, message)
        else:
            for index, boxitem in enumerate(
                self.box.boxitem_set.all().order_by("position"), start=1
            ):
                boxitem.position = index
                boxitem.verified = False
                boxitem.verified_datetime = None
                boxitem.save()
            self.box.save()
            message = (
                f"Box {self.box_identifier} has been renumber. "
                "Be sure to verify the position of each specimen."
            )
            messages.success(self.request, message)

    def add_box_item(self, **kwargs):  # noqa
        """Adds the item to the next available position in the box."""
        if self.box.status == SHIPPED:
            message = "Unable to add. Box has already been shipped."
            messages.error(self.request, message)
        else:
            try:
                box_item = BoxItem.objects.get(
                    box__box_identifier=self.box_identifier,
                    identifier=self.box_item_identifier,
                )
            except ObjectDoesNotExist:
                try:
                    box_item = BoxItem.objects.get(identifier=self.box_item_identifier)
                except ObjectDoesNotExist:
                    self.create_box_item()
                else:
                    message = format_html(
                        'Item is already packed. See box <a href="{}" '
                        'class="alert-link">{}</a>',
                        self.box_href,
                        box_item.box.human_readable_identifier,
                    )
                    messages.error(self.request, message)
            else:
                message = (
                    f"Duplicate item. {box_item.human_readable_identifier} "
                    f"is already in position {box_item.position}."
                )
                messages.error(self.request, message)

    @property
    def box_href(self):
        url_name = url_names.get(self.post_action_url)
        return reverse(
            url_name,
            kwargs={"box_identifier": self.box_identifier, "action_name": "manage"},
        )

    def create_box_item(self):
        try:
            position = self.box.next_position
        except BoxIsFullError as e:
            message = str(e)
            messages.error(self.request, message)
        else:
            box_item = BoxItem(
                box=self.box, identifier=self.box_item_identifier, position=position
            )
            box_item.save()
            if self.box.verified:
                self.box.save()



from .listboard_filter_view_mixin import ListboardFilterViewMixin
from .querystring_view_mixin import QueryStringViewMixin
from .search_form_view_mixin import SearchFormViewMixin
from .search_listboard_view_mixin import SearchListboardMixin



from __future__ import annotations

import contextlib
from copy import copy

import numpy as np
import pandas as pd
from django.apps import apps as django_apps
from django.contrib.sites.models import Site
from django.core.exceptions import FieldError
from django.db import OperationalError
from django.db.models import QuerySet
from django_crypto_fields.utils import get_encrypted_fields, has_encrypted_fields
from django_pandas.io import read_frame

from edc_constants.constants import NULL_STRING
from edc_lab.models import Panel
from edc_list_data.model_mixins import ListModelMixin, ListUuidModelMixin

from .constants import ACTION_ITEM_COLUMNS, SYSTEM_COLUMNS

__all__ = ["ModelToDataframe", "ModelToDataframeError"]


class ModelToDataframeError(Exception):
    pass


class ModelToDataframe:
    """A class to return a model or queryset as a pandas dataframe
    with custom handling for EDC models.

    For a CRF, subject_identifier and a few other columns are added.
    Custom handles edc Site and Panel FK columns, list model FK, M2M,
    etc.  If model has an FK to subject_visit adds important columns
    from the SubjetVisit model (see `add_columns_for_subject_visit`).
    Protections are in place for models or related models with
    encrypted fields.

    We are not using django_pandas read_frame since a model class
    may have M2M fields and read_frame does not handle M2M fields
    well. read_frame does not know about edc encrypted fields.

    m = ModelToDataframe(model='edc_pdutils.crf')
    my_df = m.dataframe

    See also: get_crf()
    """

    sys_field_names: tuple[str, ...] = (
        "_state",
        "_user_container_instance",
        "_domain_cache",
        "using",
        "slug",
    )
    edc_sys_columns: tuple[str, ...] = SYSTEM_COLUMNS
    action_item_columns: tuple[str, ...] = ACTION_ITEM_COLUMNS
    illegal_chars: dict[str, str] = {  # noqa: RUF012
        "\u2019": "'",
        "\u2018": "'",
        "\u201d": '"',
        "\u2013": "-",
        "\u2022": "*",
    }

    def __init__(
        self,
        model: str | None = None,
        queryset: [QuerySet] | None = None,
        query_filter: dict | None = None,
        decrypt: bool | None = None,
        drop_sys_columns: bool | None = None,
        drop_action_item_columns: bool | None = None,
        read_frame_verbose: bool | None = None,
        remove_timezone: bool | None = None,
        sites: list[int] | None = None,
    ):
        self._columns = None
        self._has_encrypted_fields = None
        self._list_columns = None
        self._encrypted_columns = None
        self._site_columns = None
        self._dataframe = pd.DataFrame()
        self.read_frame_verbose = False if read_frame_verbose is None else read_frame_verbose
        self.sites = sites
        self.drop_sys_columns = True if drop_sys_columns is None else drop_sys_columns
        self.drop_action_item_columns = (
            True if drop_action_item_columns is None else drop_action_item_columns
        )
        self.decrypt = decrypt
        self.m2m_columns = []
        self.query_filter = query_filter or {}
        self.remove_timezone = True if remove_timezone is None else remove_timezone
        if queryset:
            self.model = queryset.model._meta.label_lower
        else:
            self.model = model
        try:
            self.model_cls = django_apps.get_model(self.model)
        except LookupError as e:
            raise LookupError(f"Model is {self.model}. Got `{e}`") from e
        if self.sites:
            try:
                if queryset:
                    self.queryset = queryset.filter(site__in=sites)
                else:
                    self.queryset = self.model_cls.objects.filter(site__in=sites)
            except FieldError as e:
                if "Cannot resolve keyword 'site' into field" not in str(e):
                    raise
                self.queryset = queryset or self.model_cls.objects.all()
        else:
            self.queryset = queryset or self.model_cls.objects.all()
        # trigger query
        self.row_count = self.get_row_count()

    def get_row_count(self):
        try:
            row_count = self.queryset.count()
        except OperationalError as e:
            if "The user specified as a definer" in str(e) and self.model_cls.recreate_db_view:
                self.model_cls.recreate_db_view()
                row_count = self.queryset.count()
            else:
                raise
        return row_count

    @property
    def dataframe(self) -> pd.DataFrame:
        """Returns a pandas dataframe.

        Warning:
            If any column names collide, 'rename()' converts column
            datatype from Series to Dataframe and an error will be
            raised later on. For example, main model "visit_reason"
            and "subject_visit.visit_reason" -- in the '_dataframe',
            column "visit_reason" would become a Dataframe instead
            of a Series like all other columns.
        """
        if self._dataframe.empty and self.row_count > 0:
            if self.decrypt and self.has_encrypted_fields:
                dataframe = self.get_dataframe_with_encrypted_fields(self.row_count)
            else:
                dataframe = self.get_dataframe_without_encrypted_fields()

            dataframe = self.merge_m2ms(dataframe)

            dataframe = dataframe.rename(columns=self.columns)

            # remove timezone if asked
            if self.remove_timezone:
                for column in list(dataframe.select_dtypes(include=["datetimetz"]).columns):
                    dataframe[column] = pd.to_datetime(dataframe[column]).dt.tz_localize(None)

            # convert bool to int64
            for column in list(dataframe.select_dtypes(include=["bool"]).columns):
                dataframe[column] = (
                    dataframe[column].astype("int64").replace({True: 1, False: 0})
                )

            # convert object to str
            for column in list(dataframe.select_dtypes(include=["object"]).columns):
                dataframe[column] = dataframe[column].fillna("")
                dataframe[column] = dataframe[column].astype(str)

            # convert timedeltas to secs
            for column in list(dataframe.select_dtypes(include=["timedelta64"]).columns):
                dataframe[column] = dataframe[column].dt.total_seconds()

            # fillna
            dataframe = dataframe.fillna(value=np.nan, axis=0)

            # remove illegal chars
            for column in list(dataframe.select_dtypes(include=["object"]).columns):
                dataframe[column] = dataframe.apply(
                    lambda x, col=column: self._clean_chars(x[col]), axis=1
                )
            self._dataframe = dataframe
        return self._dataframe

    def get_dataframe_without_encrypted_fields(self) -> pd.DataFrame:
        queryset = self.queryset.values_list(*self.columns).filter(**self.query_filter)
        return pd.DataFrame(list(queryset), columns=[v for v in self.columns])

    def get_dataframe_with_encrypted_fields(self, row_count: int) -> pd.DataFrame:  # noqa: ARG002
        df = read_frame(
            self.queryset.filter(**self.query_filter), verbose=self.read_frame_verbose
        )
        return df[[col for col in self.columns]]

    def merge_m2ms(self, dataframe):
        """Merge m2m data into main dataframe.

        If m2m field name is not "name", add a class attr to
        the m2m model that returns a field_name.

        For example:

            # see edc_model_to_dataframe
            m2m_related_field = "patient_log_identifier"

        """
        for m2m_field in self.model_cls._meta.many_to_many:
            if getattr(m2m_field.related_model, "m2m_related_field", None):
                related_field = m2m_field.related_model.m2m_related_field
            else:
                related_field = "name"

            if related_field not in [
                f.name for f in m2m_field.related_model._meta.get_fields()
            ]:
                raise ModelToDataframeError(
                    f"m2m model missing `{related_field}` field. "
                    f"Parent model is {self.model_cls}. "
                    f"Got {m2m_field.related_model}. Try adding attribute "
                    "m2m_related_field={model_name: field_name} to model class "
                    f"{m2m_field.related_model}"
                )
            m2m_field_name = f"{m2m_field.name}__{related_field}"
            df_m2m = read_frame(
                self.model_cls.objects.prefetch_related(m2m_field_name)
                .filter(**{f"{m2m_field_name}__isnull": False})
                .values("id", m2m_field_name)
            )
            df_m2m = df_m2m.groupby("id")[m2m_field_name].apply(",".join).reset_index()
            df_m2m = df_m2m.rename(columns={m2m_field_name: m2m_field_name.split("__")[0]})
            dataframe = dataframe.merge(df_m2m, on="id", how="left")
        return dataframe

    def _clean_chars(self, s: str) -> str:
        if s:
            for k, v in self.illegal_chars.items():
                try:
                    s = s.replace(k, v)
                except (AttributeError, TypeError):
                    break
            return s
        return NULL_STRING

    def move_sys_columns_to_end(self, columns: dict[str, str]) -> dict[str, str]:
        system_columns = [
            f.name for f in self.model_cls._meta.get_fields() if f.name in SYSTEM_COLUMNS
        ]
        new_columns = {k: v for k, v in columns.items() if k not in system_columns}
        if (
            system_columns
            and len(new_columns.keys()) != len(columns.keys())
            and not self.drop_sys_columns
        ):
            new_columns.update({k: k for k in system_columns})
        return new_columns

    def move_action_item_columns(self, columns: dict[str, str]) -> dict[str, str]:
        action_item_columns = [
            f.name for f in self.model_cls._meta.get_fields() if f.name in ACTION_ITEM_COLUMNS
        ]
        new_columns = {k: v for k, v in columns.items() if k not in ACTION_ITEM_COLUMNS}
        if action_item_columns and (
            len(new_columns.keys()) != len(columns.keys())
            and not self.drop_action_item_columns
        ):
            new_columns.update({k: k for k in ACTION_ITEM_COLUMNS})
        return new_columns

    @property
    def has_encrypted_fields(self) -> bool:
        """Returns True if at least one field uses encryption."""
        if self._has_encrypted_fields is None:
            self._has_encrypted_fields = has_encrypted_fields(self.model_cls)
        return self._has_encrypted_fields

    @property
    def columns(self) -> dict[str, str]:
        """Return a dictionary of column names."""
        if not self._columns:
            columns_list = self.get_columns_list()
            columns = {col: col for col in columns_list}
            for column_name in columns_list:
                if column_name.endswith("_visit_id"):
                    with contextlib.suppress(FieldError):
                        columns = self.add_columns_for_subject_visit(
                            column_name=column_name, columns=columns
                        )
                if column_name.endswith("_requisition") or column_name.endswith(
                    "requisition_id"
                ):
                    columns = self.add_columns_for_subject_requisitions(columns)
            columns = self.add_columns_for_site(columns=columns)
            columns = self.add_list_model_name_columns(columns)
            columns = self.add_other_columns(columns)
            columns = self.add_subject_identifier_column(columns)
            columns = self.move_action_item_columns(columns)
            columns = self.move_sys_columns_to_end(columns)
            # ensure no encrypted fields were added
            if not self.decrypt:
                columns = {k: v for k, v in columns.items() if k not in self.encrypted_columns}
            self._columns = columns
        return self._columns

    def get_columns_list(self) -> list[str]:
        try:
            columns_list = list(self.queryset.first().__dict__.keys())
        except AttributeError as e:
            if "__dict__" in str(e):
                columns_list = list(self.queryset._fields)
            else:
                raise
        for name in self.sys_field_names:
            with contextlib.suppress(ValueError):
                columns_list.remove(name)
        if not self.decrypt:
            columns_list = [col for col in columns_list if col not in self.encrypted_columns]
        return columns_list

    @property
    def encrypted_columns(self) -> list[str]:
        """Return a sorted list of column names that use encryption."""
        if not self._encrypted_columns:
            self._encrypted_columns = [
                field.name for field in get_encrypted_fields(self.model_cls)
            ]
            self._encrypted_columns = list(set(self._encrypted_columns))
            self._encrypted_columns.sort()
        return self._encrypted_columns

    @property
    def list_columns(self) -> list[str]:
        """Return a list of column names with fk to a list model."""

        if not self._list_columns:
            list_columns = []
            for fld_cls in self.model_cls._meta.get_fields():
                if (
                    hasattr(fld_cls, "related_model")
                    and fld_cls.related_model
                    and issubclass(fld_cls.related_model, (ListModelMixin, ListUuidModelMixin))
                ):
                    list_columns.append(fld_cls.attname)  # noqa: PERF401
            self._list_columns = list(set(list_columns))
        return self._list_columns

    @property
    def site_columns(self) -> list[str]:
        """Return a list of column names with fk to a site model."""

        if not self._site_columns:
            site_columns = []
            for fld_cls in self.model_cls._meta.get_fields():
                if (
                    hasattr(fld_cls, "related_model")
                    and fld_cls.related_model
                    and issubclass(fld_cls.related_model, (Site,))
                ):
                    site_columns.append(fld_cls.attname)
            self._site_columns = list(set(site_columns))
        return self._site_columns

    @property
    def other_columns(self) -> list[str]:
        """Return other column names with fk to a common models."""
        related_model = [Site, Panel]
        if not self._list_columns:
            list_columns = []
            for fld_cls in self.model_cls._meta.get_fields():
                if (
                    hasattr(fld_cls, "related_model")
                    and fld_cls.related_model
                    and fld_cls.related_model in related_model
                ):
                    list_columns.append(fld_cls.attname)
            self._list_columns = list(set(list_columns))
        return self._list_columns

    def add_subject_identifier_column(self, columns: dict[str, str]) -> dict[str, str]:
        if "subject_identifier" not in [v for v in columns.values()]:
            subject_identifier_column = None
            id_columns = [col.replace("_id", "") for col in columns if col.endswith("_id")]
            for col in id_columns:
                field = getattr(self.model_cls, col, None)
                if field and [
                    fld.name
                    for fld in field.field.related_model._meta.get_fields()
                    if fld.name == "subject_identifier"
                ]:
                    subject_identifier_column = f"{col}__subject_identifier"
                    break
            if subject_identifier_column:
                columns.update({subject_identifier_column: "subject_identifier"})
        return columns

    @staticmethod
    def add_columns_for_subject_visit(
        column_name: str = None, columns: dict[str, str] = None
    ) -> dict[str, str]:
        if "subject_identifier" not in [v for v in columns.values()]:
            columns.update(
                {f"{column_name}__appointment__subject_identifier": "subject_identifier"}
            )
        columns.update({f"{column_name}__appointment__appt_datetime": "appointment_datetime"})
        columns.update({f"{column_name}__appointment__visit_code": "visit_code"})
        columns.update(
            {f"{column_name}__appointment__visit_code_sequence": "visit_code_sequence"}
        )
        columns.update({f"{column_name}__report_datetime": "visit_datetime"})
        columns.update({f"{column_name}__reason": "visit_reason"})
        return columns

    @staticmethod
    def add_columns_for_subject_requisitions(
        columns: dict[str, str] = None,
    ) -> dict[str, str]:
        for col in copy(columns):
            if col.endswith("_requisition_id"):
                col_prefix = col.split("_")[0]
                column_name = col.split("_id")[0]
                columns.update(
                    {
                        f"{column_name}__requisition_identifier": (
                            f"{col_prefix}_requisition_identifier"
                        )
                    }
                )
                columns.update(
                    {f"{column_name}__drawn_datetime": f"{col_prefix}_drawn_datetime"}
                )
                columns.update({f"{column_name}__is_drawn": f"{col_prefix}_is_drawn"})
        return columns

    def add_columns_for_site(self, columns: dict[str, str] = None) -> dict[str, str]:
        for col in copy(columns):
            if col in self.site_columns:
                col_prefix = col.split("_id")[0]
                columns.update({f"{col_prefix}__name": f"{col_prefix}_name"})
        return columns

    def add_list_model_name_columns(self, columns: dict[str, str] = None) -> dict[str, str]:
        for col in copy(columns):
            if col in self.list_columns:
                column_name = col.split("_id")[0]
                columns.update({f"{column_name}__name": f"{column_name}_name"})
        return columns

    def add_other_columns(self, columns: dict[str, str] = None) -> dict[str, str]:
        for col in copy(columns):
            if col in self.other_columns:
                column_name = col.split("_id")[0]
                columns.update({f"{column_name}__name": f"{column_name}_name"})
        return columns



from django.conf import settings
from django.core.validators import MaxValueValidator, MinValueValidator
from django.db import models
from django.utils.html import format_html

from edc_constants.choices import YES_NO
from edc_constants.constants import YES
from edc_model import models as edc_models

from .choices import MISSED_PILLS


class MedicationAdherenceModelMixin(models.Model):
    visual_score_slider = models.CharField(
        verbose_name="Visual score", max_length=3, help_text="%"
    )

    visual_score_confirmed = models.IntegerField(
        verbose_name=format_html(
            "<B><font color='{}'>{}</font></B>: {}.",
            "orange",
            "Interviewer",
            "please transcribe the score indicated from above",
        ),
        validators=[MinValueValidator(0), MaxValueValidator(100)],
        help_text="%",
    )

    last_missed_pill = models.CharField(
        verbose_name="When was the last time you missed your study pill?",
        max_length=25,
        choices=MISSED_PILLS,
    )

    pill_count_performed = models.CharField(
        verbose_name="Was a pill count performed",
        max_length=5,
        choices=YES_NO,
        default=YES,
    )

    pill_count_not_performed_reason = models.TextField(
        verbose_name="If not performed, please explain ...",
        default="",
        blank=True,
    )

    pill_count = models.IntegerField(
        verbose_name="Number of pills left in the bottle", null=True, blank=True
    )

    missed_pill_reason = models.ManyToManyField(
        f"{settings.LIST_MODEL_APP_LABEL}.NonAdherenceReasons",
        verbose_name="Reasons for missing study pills",
        blank=True,
    )

    other_missed_pill_reason = edc_models.OtherCharField(max_length=150)

    class Meta:
        abstract = True
        verbose_name = "Medication Adherence"
        verbose_name_plural = "Medication Adherence"



action_fields = (
    "action_identifier",
    "action_item",
    "parent_action_item",
    "related_action_item",
)

action_fieldset_tuple = ("Action", {"classes": ("collapse",), "fields": action_fields})



from django.contrib import admin
from django_audit_fields.admin import audit_fieldset_tuple

from edc_model_admin.history import SimpleHistoryAdmin

from ...admin_site import edc_pharmacy_admin
from ...forms import ProductForm
from ...models import Product
from ..model_admin_mixin import ModelAdminMixin
from ..remove_fields_for_blinded_users import remove_fields_for_blinded_users


@admin.register(Product, site=edc_pharmacy_admin)
class ProductAdmin(ModelAdminMixin, SimpleHistoryAdmin):
    change_list_title = "Pharmacy: Products"
    history_list_display = ()
    show_object_tools = True
    show_cancel = True
    list_per_page = 20

    form = ProductForm

    fieldsets = (
        (
            None,
            {
                "fields": (
                    [
                        "product_identifier",
                        "formulation",
                        "assignment",
                        "name",
                    ]
                )
            },
        ),
        audit_fieldset_tuple,
    )

    list_display = (
        "identifier",
        "name",
        "formulation",
        "assignment",
        "created",
        "modified",
    )
    list_filter = (
        "formulation",
        "assignment",
    )
    search_fields = (
        "product_identifier",
        "lot__lot_no",
    )
    ordering = ("product_identifier",)
    readonly_fields = ("product_identifier", "name")
    radio_fields = {"assignment": admin.VERTICAL}

    @admin.display(description="PRODUCT #", ordering="product_identifier")
    def identifier(self, obj):
        return obj.product_identifier.split("-")[0]

    def get_readonly_fields(self, request, obj=None):
        if obj:
            return self.readonly_fields + ("formulation", "assignment")
        return self.readonly_fields

    def get_list_display(self, request):
        fields = super().get_list_display(request)
        fields = remove_fields_for_blinded_users(request, fields)
        return fields

    def get_list_filter(self, request):
        fields = super().get_list_filter(request)
        fields = remove_fields_for_blinded_users(request, fields)
        return fields

    def get_search_fields(self, request):
        fields = super().get_search_fields(request)
        fields = remove_fields_for_blinded_users(request, fields)
        return fields



from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ..models import StockRequest


def stock_request_status_counts(obj: StockRequest) -> dict[str, int]:
    return dict(
        total=obj.stockrequestitem_set.all().count(),
        pending=obj.stockrequestitem_set.filter(allocation__isnull=True).count(),
        allocated=(obj.stockrequestitem_set.filter(allocation__isnull=False).count()),
        allocated_and_transferred=obj.stockrequestitem_set.filter(
            allocation__isnull=False, allocation__stock__location=obj.location
        ).count(),
    )



# Generated by Django 5.2.6 on 2025-09-18 00:24

import django.utils.timezone
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("edc_export", "0022_alter_datarequest_description_and_more"),
    ]

    operations = [
        migrations.AlterField(
            model_name="datarequesthistory",
            name="archive_filename",
            field=models.CharField(default="", max_length=200),
        ),
        migrations.AlterField(
            model_name="datarequesthistory",
            name="emailed_to",
            field=models.EmailField(default="", max_length=254),
        ),
        migrations.AlterField(
            model_name="datarequesthistory",
            name="exported_datetime",
            field=models.DateTimeField(default=django.utils.timezone.now),
        ),
    ]



import pandas as pd

from ..constants import ACTION_ITEM_COLUMNS, SYSTEM_COLUMNS
from ..dialects import CrfDialect
from .df_handler import DfHandler


class CrfDfHandlerError(Exception):
    pass


class CrfDfHandler(DfHandler):
    crf_dialect_cls = CrfDialect
    # visit_tbl = None
    visit_column: str = "subject_visit_id"
    appointment_tbl: str = "edc_appointment_appointment"
    registered_subject_tbl: str = "edc_registration_registeredsubject"
    dialect_select_visit_and_related: str = "select_visit_and_related"

    action_item_columns = ACTION_ITEM_COLUMNS
    system_columns = SYSTEM_COLUMNS
    sort_by = ["subject_identifier", "visit_datetime"]
    exclude_export_columns = True
    exclude_system_columns = False
    exclude_columns = ["form_as_json"]

    def __init__(self, exclude_system_columns=None, **kwargs):
        self._df_visit_and_related = pd.DataFrame()
        self.crf_dialect = self.crf_dialect_cls(self)
        self.exclude_system_columns = exclude_system_columns or self.exclude_system_columns
        super().__init__(**kwargs)

    def prepare_dataframe(self, **kwargs):
        """Merges CRF dataframe with df_visit_and_related
        on visit_column.
        """
        if not self.visit_column:
            raise CrfDfHandlerError(
                f"visit column cannot be None. See class attr 'visit_column' on {self!r}"
            )
        self.dataframe = pd.merge(
            left=self.dataframe,
            right=self.df_visit_and_related,
            how="left",
            on=self.visit_column,
            suffixes=("_notused", ""),
        )
        self.dataframe = self.dataframe[self.columns]

    @property
    def columns(self):
        """Returns a list of column names."""
        crf_columns = list(self.dataframe.columns)
        crf_columns.pop(crf_columns.index(self.visit_column))
        columns = list(self.df_visit_and_related.columns)
        columns.extend([c for c in crf_columns if c not in columns])
        columns = [col for col in columns if col not in self.exclude_columns]
        # "export_" columns
        if self.exclude_export_columns:
            columns = [col for col in columns if not col.startswith("export_")]
        if self.exclude_system_columns:
            columns = [col for col in columns if col not in self.system_columns]
        else:
            if [col for col in columns if col in self.action_item_columns]:
                # "action_item" columns, move to the end
                columns = [col for col in columns if col not in self.action_item_columns]
                columns.extend(self.action_item_columns)
            # "system" columns, move to the end
            columns = [col for col in columns if col not in self.system_columns]
            columns.extend(self.system_columns)
        return columns

    @property
    def df_visit_and_related(self):
        """Returns a dataframe of the crf_dialect's `select_visit_and_related`
        SQL statement.
        """
        if self._df_visit_and_related.empty:
            sql, params = getattr(self.crf_dialect, self.dialect_select_visit_and_related)
            self._df_visit_and_related = self.db.read_sql(sql, params=params)
        return self._df_visit_and_related



# Generated by Django 2.2.3 on 2019-09-17 18:33

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [("edc_data_manager", "0005_auto_20190826_1646")]

    operations = [
        migrations.AlterField(
            model_name="datadictionary",
            name="help_text",
            field=models.TextField(null=True),
        ),
        migrations.AlterField(
            model_name="historicaldatadictionary",
            name="help_text",
            field=models.TextField(null=True),
        ),
    ]



from django.conf import settings

from edc_auth.constants import AUDITOR_ROLE, CLINICIAN_ROLE, CLINICIAN_SUPER_ROLE
from edc_auth.site_auths import site_auths

from .auth_objects import (
    PROTOCOL_INCIDENT_VIEW,
    PROTOCOL_VIOLATION,
    PROTOCOL_VIOLATION_VIEW,
    protocol_incident_codenames,
    protocol_incident_view_codenames,
    protocol_violation_codenames,
    protocol_violation_view_codenames,
)
from .constants import PROTOCOL_DEVIATION_VIOLATION, PROTOCOL_INCIDENT


def update_site_auths() -> None:
    incident_type = getattr(
        settings, "EDC_PROTOCOL_VIOLATION_TYPE", PROTOCOL_DEVIATION_VIOLATION
    )

    site_auths.add_group(*protocol_violation_codenames, name=PROTOCOL_VIOLATION)
    site_auths.add_group(*protocol_violation_view_codenames, name=PROTOCOL_VIOLATION_VIEW)
    site_auths.add_group(*protocol_incident_codenames, name=PROTOCOL_INCIDENT)
    site_auths.add_group(*protocol_incident_view_codenames, name=PROTOCOL_INCIDENT_VIEW)

    if incident_type == PROTOCOL_DEVIATION_VIOLATION:
        site_auths.update_role(PROTOCOL_VIOLATION, name=CLINICIAN_ROLE)
        site_auths.update_role(PROTOCOL_VIOLATION, name=CLINICIAN_SUPER_ROLE)
        site_auths.update_role(PROTOCOL_VIOLATION_VIEW, name=AUDITOR_ROLE)
    elif incident_type == PROTOCOL_INCIDENT:
        site_auths.update_role(PROTOCOL_INCIDENT, name=CLINICIAN_ROLE)
        site_auths.update_role(PROTOCOL_INCIDENT, name=CLINICIAN_SUPER_ROLE)
        site_auths.update_role(PROTOCOL_INCIDENT_VIEW, name=AUDITOR_ROLE)
    else:
        raise ValueError(
            "Invalid value for settings.EDC_PROTOCOL_VIOLATION_TYPE. "
            f"Expected `{PROTOCOL_INCIDENT}` or `{PROTOCOL_DEVIATION_VIOLATION}`. "
            f"Got {incident_type}."
        )


update_site_auths()



from edc_constants.constants import YES
from edc_form_validators import FormValidator

from .ltfu_form_validator_mixin import LtfuFormValidatorMixin


class LtfuFormValidator(LtfuFormValidatorMixin, FormValidator):
    def clean(self):
        self.validate_ltfu()
        self.required_if(YES, field="home_visit", field_required="home_visit_detail")
        self.validate_other_specify(
            field="loss_category", other_specify_field="loss_category_other"
        )



from .blood_pressure_form_validator_mixin import BloodPressureFormValidatorMixin
from .bmi_form_validator_mixin import BmiFormValidatorMixin
from .weight_height_with_bmi_form_validator_mixin import (
    WeightHeightBmiFormValidatorMixin,
)



from __future__ import annotations

from urllib.parse import urlencode

from django.contrib import messages
from django.http.response import HttpResponseRedirect
from django.urls import reverse
from django.utils.encoding import force_str

from edc_dashboard.url_names import InvalidDashboardUrlName, url_names


class ModelAdminRedirectOnDeleteMixin:
    """A mixin to redirect post delete.

    If `post_url_on_delete_name` and `post_full_url_on_delete` are
    not set, does nothing.
    """

    post_url_on_delete_name = "subject_dashboard_url"  # lookup key for url_names dict
    post_full_url_on_delete = None

    def __init__(self, *args):
        self.post_url_on_delete = None
        super().__init__(*args)

    def get_post_url_on_delete_name(self, request) -> str:
        return url_names.get(self.post_url_on_delete_name)

    def get_post_full_url_on_delete(self, request) -> str | None:
        return self.post_full_url_on_delete

    def get_post_url_on_delete(self, request, obj) -> str | None:
        """Returns a url for the redirect after delete."""
        post_url_on_delete = None
        querystring = urlencode(self.post_url_on_delete_querystring_kwargs(request, obj))
        if self.get_post_full_url_on_delete(request):
            post_url_on_delete = reverse(self.get_post_full_url_on_delete(request))
        else:
            try:
                url_name = self.get_post_url_on_delete_name(request)
            except InvalidDashboardUrlName:
                if self.post_url_on_delete_name:
                    raise
                url_name = None
            if url_name:
                kwargs = self.post_url_on_delete_kwargs(request, obj)
                post_url_on_delete = reverse(url_name, kwargs=kwargs)
        if post_url_on_delete and querystring:
            post_url_on_delete = f"{post_url_on_delete}?{querystring}"
        return post_url_on_delete

    def post_url_on_delete_kwargs(self, request, obj) -> dict:
        """Returns kwargs needed to reverse the url.

        Override.
        """
        return {}

    def post_url_on_delete_querystring_kwargs(self, request, obj) -> dict:
        """Returns kwargs for a querystring for the reversed url.

        Override.
        """
        return {}

    def delete_model(self, request, obj) -> None:
        """Overridden to intercept the obj to reverse
        the post_url_on_delete
        """
        self.post_url_on_delete = self.get_post_url_on_delete(request, obj)
        obj.delete()

    def response_delete(self, request, obj_display, obj_id):
        """Overridden to redirect to `post_url_on_delete`, if not None."""
        if self.post_url_on_delete:
            opts = self.model._meta
            msg = 'The %(name)s "%(obj)s" was deleted successfully.' % {
                "name": force_str(opts.verbose_name),
                "obj": force_str(obj_display),
            }
            messages.add_message(request, messages.SUCCESS, msg)
            return HttpResponseRedirect(self.post_url_on_delete)
        return super().response_delete(request, obj_display, obj_id)



from __future__ import annotations

from django.utils.translation import gettext as _

from edc_model_admin.list_filters import (
    ListFieldWithOtherListFilter,
    PastDateListFilter,
)


class AeAwarenessListFilter(PastDateListFilter):
    title = _("Awareness date")

    parameter_name = "ae_awareness_date"
    field_name = "ae_awareness_date"


class DeathDateListFilter(PastDateListFilter):
    title = _("Death date")

    parameter_name = "death_datetime"
    field_name = "death_datetime"


class CauseOfDeathListFilter(ListFieldWithOtherListFilter):
    title = "Cause of death"
    parameter_name = "cause_of_death"
    other_parameter_name = "cause_of_death_other"


class AeClassificationListFilter(ListFieldWithOtherListFilter):
    title = "Classification"
    parameter_name = "ae_classification"
    other_parameter_name = "ae_classification_other"



# Generated by Django 4.2.11 on 2024-08-14 15:55

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("edc_qareports", "0011_auto_20240813_0219"),
    ]

    operations = [
        migrations.AlterField(
            model_name="qareportnote",
            name="note",
            field=models.TextField(blank=True, null=True),
        ),
    ]



from django.db import models

from edc_model.models import BaseUuidModel


class ZplLabelTemplates(BaseUuidModel):
    name = models.CharField(max_length=50, unique=True)

    zpl_data = models.TextField(max_length=1000)

    def __str__(self):
        return self.name

    class Meta(BaseUuidModel.Meta):
        verbose_name = "Label template"
        verbose_name_plural = "Label templates"



import sys
from copy import deepcopy

from django.apps import apps as django_apps
from django.core.management.color import color_style
from django.utils.module_loading import import_module, module_has_submodule


class AlreadyRegistered(Exception):
    pass


class SiteDataManagerError(Exception):
    pass


class SiteDataManager:
    def __init__(self):
        self.registry = {}
        self.loaded = False

    def register(self, rule_handler=None):
        if "makemigrations" not in sys.argv:
            self._register(rule_handler=rule_handler)

    def _register(self, rule_handler=None):
        if rule_handler.name in self.registry:
            raise AlreadyRegistered(
                f"Query rule_handler already registered. Got {rule_handler}."
            )
        self.registry.update({rule_handler.name: rule_handler})

    def get_rule_handler(self, name):
        name = name or "default"
        if name not in self.registry:
            raise SiteDataManagerError(f"Query rule handler is not registered. Got {name}.")
        return self.registry.get(name)

    def get_rule_handlers(self, model_name=None):
        rule_handlers = []
        if model_name:
            for k, v in self.registry.items():
                if v.model_name == model_name:
                    rule_handlers.append(v)
        else:
            return list(self.registry.values())
        return rule_handlers

    def autodiscover(self, module_name=None, verbose=True):
        """Autodiscovers query rule classes in the data_manager.py file of
        any INSTALLED_APP.
        """
        module_name = module_name or "data_manager"
        writer = sys.stdout.write if verbose else lambda x: x
        style = color_style()
        writer(f" * checking for data manager {module_name} ...\n")
        for app in django_apps.app_configs:
            writer(f" * searching {app}           \r")
            try:
                mod = import_module(app)
                try:
                    before_import_registry = deepcopy(site_data_manager.registry)
                    import_module(f"{app}.{module_name}")
                    writer(f"   - registered '{module_name}' from '{app}'\n")
                except SiteDataManagerError as e:
                    writer(f"   - loading {app}.{module_name} ... ")
                    writer(style.ERROR(f"ERROR! {e}\n"))
                except ImportError as e:
                    site_data_manager.registry = before_import_registry
                    if module_has_submodule(mod, module_name):
                        raise SiteDataManagerError(str(e))
            except ImportError:
                pass


site_data_manager = SiteDataManager()



# Generated by Django 2.1 on 2018-09-21 20:42

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [("edc_pdutils", "0001_initial")]

    operations = [
        migrations.RemoveField(model_name="datarequesthistory", name="data_request"),
        migrations.RemoveField(model_name="historicaldatarequest", name="history_user"),
        migrations.DeleteModel(name="DataRequest"),
        migrations.DeleteModel(name="DataRequestHistory"),
        migrations.DeleteModel(name="HistoricalDataRequest"),
    ]



from django.conf import settings

from .base_form_validator import BaseFormValidator


class LocaleValidator(BaseFormValidator):
    def __init__(self, locale: str | None = None, **kwargs):
        self.locale = locale or settings.LANGUAGE_CODE
        super().__init__(**kwargs)



from copy import copy

from edc_navbar import Navbar, NavbarItem, site_navbars

specimens_navbar = NavbarItem(
    name="specimens",
    label="Specimens",
    title="Specimens",
    fa_icon="fa-flask",
    codename="edc_lab_dashboard.nav_lab_section",
    url_name="requisition_listboard_url",
)

_specimens_navbar = copy(specimens_navbar)
_specimens_navbar.active = True
_specimens_navbar.label = None

navbar = Navbar(name="specimens")

navbar.register(
    NavbarItem(
        name="requisition",
        label="Requisition",
        codename="edc_lab_dashboard.nav_lab_requisition",
        url_name="requisition_listboard_url",
    )
)

navbar.register(
    NavbarItem(
        name="receive",
        label="Receive",
        codename="edc_lab_dashboard.nav_lab_receive",
        url_name="receive_listboard_url",
    )
)

navbar.register(
    NavbarItem(
        name="process",
        label="Process",
        codename="edc_lab_dashboard.nav_lab_process",
        url_name="process_listboard_url",
    )
)

navbar.register(
    NavbarItem(
        name="pack",
        label="Pack",
        codename="edc_lab_dashboard.nav_lab_pack",
        url_name="pack_listboard_url",
    )
)

navbar.register(
    NavbarItem(
        name="manifest",
        label="Manifest",
        codename="edc_lab_dashboard.nav_lab_manifest",
        url_name="manifest_listboard_url",
    )
)

navbar.register(
    NavbarItem(
        name="aliquot",
        label="Aliquot",
        codename="edc_lab_dashboard.nav_lab_aliquot",
        url_name="aliquot_listboard_url",
    )
)

navbar.register(
    NavbarItem(
        name="result",
        label="Result",
        codename="edc_lab_dashboard.nav_lab_result",
        url_name="result_listboard_url",
    )
)

navbar.register(
    NavbarItem(
        name="specimens",
        title="Specimens",
        fa_icon="fa-flask",
        codename="edc_lab_dashboard.nav_lab_section",
        url_name="requisition_listboard_url",
        active=True,
    )
)

site_navbars.register(navbar)



from __future__ import annotations

import calendar
import sys
import warnings
from datetime import datetime
from typing import TYPE_CHECKING, Any

from dateutil.relativedelta import relativedelta
from django.apps import apps as django_apps
from django.conf import settings
from django.contrib import messages
from django.contrib.admin.utils import NotRelationField, get_model_from_relation
from django.contrib.messages import ERROR, SUCCESS
from django.core.exceptions import (
    ImproperlyConfigured,
    ObjectDoesNotExist,
    ValidationError,
)
from django.core.handlers.wsgi import WSGIRequest
from django.db import IntegrityError, transaction
from django.db.models import Count, ProtectedError
from django.urls import reverse
from django.utils.translation import gettext as _

from edc_constants.constants import CLINIC, NOT_APPLICABLE, OK
from edc_constants.constants import ERROR as ERROR_CODE
from edc_dashboard.url_names import url_names
from edc_form_validators import INVALID_ERROR
from edc_metadata.constants import CRF, REQUIRED, REQUISITION
from edc_metadata.utils import (
    get_crf_metadata_model_cls,
    get_requisition_metadata_model_cls,
    has_keyed_metadata,
)
from edc_utils import convert_php_dateformat
from edc_utils.date import to_local
from edc_visit_schedule.exceptions import (
    ScheduledVisitWindowError,
    UnScheduledVisitWindowError,
)
from edc_visit_schedule.site_visit_schedules import site_visit_schedules
from edc_visit_schedule.utils import get_default_max_visit_window_gap, is_baseline
from edc_visit_tracking.utils import get_allow_missed_unscheduled_appts

from .choices import DEFAULT_APPT_REASON_CHOICES
from .constants import (
    CANCELLED_APPT,
    COMPLETE_APPT,
    INCOMPLETE_APPT,
    MISSED_APPT,
    NEW_APPT,
    SCHEDULED_APPT,
    SKIPPED_APPT,
    UNSCHEDULED_APPT,
)
from .exceptions import (
    AppointmentBaselineError,
    AppointmentDatetimeError,
    AppointmentMissingValuesError,
    AppointmentWindowError,
    UnscheduledAppointmentError,
)

if TYPE_CHECKING:
    from decimal import Decimal

    from django.db.models import QuerySet

    from edc_crf.model_mixins import CrfModelMixin as Base
    from edc_metadata.model_mixins.creates import CreatesMetadataModelMixin

    from .models import Appointment, AppointmentType

    class RelatedVisitModel(CreatesMetadataModelMixin, Base):
        appointment: Appointment


class AppointmentDateWindowPeriodGapError(Exception):
    pass


class AppointmentAlreadyStarted(Exception):  # noqa: N818
    pass


def get_appointment_model_name() -> str:
    return "edc_appointment.appointment"


def get_appointment_model_cls() -> type[Appointment]:
    return django_apps.get_model(get_appointment_model_name())


def get_appointment_type_model_name() -> str:
    return "edc_appointment.appointmenttype"


def get_appointment_type_model_cls() -> type[AppointmentType]:
    return django_apps.get_model(get_appointment_type_model_name())


def get_allow_skipped_appt_using() -> dict[str, tuple[str, str]]:
    """Return dict from settings or empty dictionary.

    For example:
         = {"my_app.crfone": ("next_appt_date", "next_visit_code")}
    """
    dct = getattr(settings, "EDC_APPOINTMENT_ALLOW_SKIPPED_APPT_USING", {})
    if len(dct.keys()) > 1:
        raise ImproperlyConfigured(
            "Only one model may be specified. See "
            f"settings.EDC_APPOINTMENT_ALLOW_SKIPPED_APPT_USING. Got {dct}."
        )
    return dct


def get_max_months_to_next_appointment():
    return getattr(settings, "EDC_APPOINTMENT_MAX_MONTHS_TO_NEXT_APPT", 6)


def allow_clinic_on_weekend():
    return getattr(settings, "EDC_APPOINTMENT_ALLOW_CLINIC_ON_WEEKENDS", False)


def get_max_months_to_next_appointment_as_rdelta():
    max_months = get_max_months_to_next_appointment()
    return relativedelta(months=max_months)


def raise_on_appt_may_not_be_missed(
    appointment: Appointment = None,
    appt_timing: str | None = None,
):
    if appointment.id:
        appt_timing = appt_timing or appointment.appt_timing
        if (
            appt_timing
            and appt_timing in [MISSED_APPT, NOT_APPLICABLE]
            and is_baseline(instance=appointment)
        ):
            errmsg = "Invalid. A baseline appointment may not be reported as missed"
            if appt_timing == NOT_APPLICABLE:
                errmsg = "This field is applicable"
            raise AppointmentBaselineError(errmsg)
        if (
            appointment.visit_code_sequence is not None
            and appt_timing
            and appointment.visit_code_sequence > 0
            and appt_timing == MISSED_APPT
            and not get_allow_missed_unscheduled_appts()
        ):
            raise UnscheduledAppointmentError(
                "Invalid. An unscheduled appointment may not be reported as missed."
                "Try to cancel the appointment instead. "
            )


def get_appointment_form_meta_options() -> dict:
    return getattr(
        settings, "EDC_APPOINTMENT_FORM_META_OPTIONS", dict(labels={}, help_texts={})
    )


def get_appt_reason_choices() -> tuple[str, ...]:
    """Returns a customized tuple of choices otherwise the default.

    Note: You can only change the right side of any tuple.

    For example:
      ((SCHEDULED_APPT, "some custom text"),
        (UNSCHEDULED_APPT, "some custom text"))

    See also: formfield_for_choice_field in modeladmin class.
    """

    settings_attr = "EDC_APPOINTMENT_APPT_REASON_CHOICES"
    appt_reason_choices = getattr(settings, settings_attr, DEFAULT_APPT_REASON_CHOICES)
    keys = sorted([choice[0] for choice in appt_reason_choices])
    if SCHEDULED_APPT not in keys or UNSCHEDULED_APPT not in keys:
        raise ImproperlyConfigured(
            "Invalid value for EDC_APPOINTMENT_APPT_REASON_CHOICES. "
            f"Expected a choices tuple with keys `{SCHEDULED_APPT}` and `{UNSCHEDULED_APPT}`. "
            f"See {settings_attr}."
        )
    return appt_reason_choices


def get_appt_type_default() -> str:
    """Returns the default appointment type name."""
    return getattr(settings, "EDC_APPOINTMENT_APPT_TYPE_DEFAULT", CLINIC)


def get_appt_reason_default() -> str:
    """Returns the default appointment reason."""
    value = getattr(settings, "EDC_APPOINTMENT_APPT_REASON_DEFAULT", None)
    if not value:
        value = getattr(settings, "EDC_APPOINTMENT_DEFAULT_APPT_REASON", None)
        warnings.warn(
            "Settings attribute `EDC_APPOINTMENT_DEFAULT_APPT_REASON` has "
            "been deprecated in favor of `EDC_APPOINTMENT_APPT_REASON_DEFAULT`. ",
            DeprecationWarning,
        )
    return value or SCHEDULED_APPT


def cancelled_appointment(appointment: Appointment) -> None:
    """Try to delete subject visit if appt status = CANCELLED_APPT"""
    try:
        cancelled = appointment.appt_status == CANCELLED_APPT
    except AttributeError as e:
        if "appt_status" not in str(e):
            raise
    else:
        if (
            cancelled
            and appointment.visit_code_sequence > 0
            and "historical" not in appointment._meta.label_lower
            and not appointment.crf_metadata_keyed_exists
            and not appointment.requisition_metadata_keyed_exists
        ):
            try:
                related_visit = appointment.related_visit_model_cls().objects.get(
                    appointment=appointment
                )
            except ObjectDoesNotExist:
                appointment.delete()
            else:
                with transaction.atomic():
                    try:
                        related_visit.delete()
                    except ProtectedError:
                        pass
                    else:
                        appointment.delete()


def missed_appointment(appointment: Appointment) -> None:
    """Try to create_missed_visit_from_appointment if
    appt_status == missed.
    """
    try:
        missed = appointment.appt_timing == MISSED_APPT
    except AttributeError as e:
        if "appt_timing" not in str(e):
            raise
    else:
        if (
            missed
            and appointment.visit_code_sequence == 0
            and "historical" not in appointment._meta.label_lower
        ):
            try:
                appointment.create_missed_visit_from_appointment()
            except AttributeError as e:
                if "create_missed_visit" not in str(e):
                    raise


def reset_visit_code_sequence_or_pass(
    subject_identifier: str = None,
    visit_schedule_name: str = None,
    schedule_name: str = None,
    visit_code: str = None,
    appointment: Appointment | None = None,
    write_stdout: bool | None = None,
) -> Appointment | None:
    """Validate the order of the appointment visit code sequences
    relative to the appt_datetime and reset the visit code sequences
    if needed.

    Delete and recreate metadata

    Also do the same for the `related_visit`, if it exists.
    """
    opts = dict(
        subject_identifier=subject_identifier,
        visit_schedule_name=visit_schedule_name,
        schedule_name=schedule_name,
        visit_code=visit_code,
    )
    qs = get_appointment_model_cls().objects.filter(**opts).order_by("appt_datetime")
    expected = list(range(0, qs.count()))
    actual = [o.visit_code_sequence for o in qs]
    if actual != expected:
        if write_stdout:
            sys.stdout.write(
                "     - Resetting for "
                f"{subject_identifier} {visit_code}: {actual=} {expected=} ...\n"
            )
        # reset visit code sequence for this visit code
        get_crf_metadata_model_cls().objects.filter(visit_code_sequence__gt=0, **opts).delete()
        get_requisition_metadata_model_cls().objects.filter(
            visit_code_sequence__gt=0, **opts
        ).delete()

        try:
            with transaction.atomic():
                # set appt and related visit visit_code_sequences to the
                # negative of the current value
                for obj in get_appointment_model_cls().objects.filter(
                    visit_code_sequence__gt=0, **opts
                ):
                    obj.visit_code_sequence = obj.visit_code_sequence * -1
                    obj.save_base(update_fields=["visit_code_sequence"])
                    if getattr(obj, "related_visit", None):
                        obj.related_visit.visit_code_sequence = obj.visit_code_sequence
                        obj.related_visit.save_base(update_fields=["visit_code_sequence"])
                        obj.related_visit.metadata_create()

                # reset sequence order by appt_datetime
                for index, obj in enumerate(
                    get_appointment_model_cls()
                    .objects.filter(visit_code_sequence__lt=0, **opts)
                    .order_by("appt_datetime"),
                    start=1,
                ):
                    obj.visit_code_sequence = index
                    obj.save_base(update_fields=["visit_code_sequence"])
                    if getattr(obj, "related_visit", None):
                        obj.related_visit.visit_code_sequence = index
                        obj.related_visit.save_base(update_fields=["visit_code_sequence"])
                        obj.related_visit.metadata_create()

        except IntegrityError:
            raise
        if appointment:
            # refresh the given appt if not None since
            # appointment visit_code_sequence may have changed
            appointment = get_appointment_model_cls().objects.get(id=appointment.id)
    return appointment


def reset_visit_code_sequence_for_subject(
    subject_identifier: str = None,
    visit_schedule_name: str = None,
    schedule_name: str = None,
) -> None:
    """Resets / validates appointment `visit code sequences` for any
    `visit code` with unscheduled appointments for the given subject
     and schedule.

     Wrapper for function `reset_visit_code_sequence_or_pass`.
    """
    ann = (
        get_appointment_model_cls()
        .objects.values("visit_code")
        .filter(subject_identifier=subject_identifier, visit_code_sequence__gt=0)
        .annotate(Count("visit_code"))
    )
    for visit_code in [obj.get("visit_code") for obj in ann]:
        reset_visit_code_sequence_or_pass(
            subject_identifier=subject_identifier,
            visit_schedule_name=visit_schedule_name,
            schedule_name=schedule_name,
            visit_code=visit_code,
        )


def delete_appointment_in_sequence(appointment: Any, from_post_delete=None) -> None:
    if not from_post_delete:
        with transaction.atomic():
            appointment.delete()
        reset_visit_code_sequence_or_pass(
            subject_identifier=appointment.subject_identifier,
            visit_schedule_name=appointment.visit_schedule_name,
            schedule_name=appointment.schedule_name,
            visit_code=appointment.visit_code,
        )


def update_appt_status(appointment: Appointment, save: bool | None = None):
    """Sets appt_status, and if save is True, calls save_base().

    This is useful if checking `appt_status` is correct
    relative to the visit tracking model and CRFs and
    requisitions
    """
    if appointment.appt_status == CANCELLED_APPT or appointment.appt_status == SKIPPED_APPT:
        pass
    elif not appointment.related_visit:
        appointment.appt_status = NEW_APPT
    elif (
        appointment.crf_metadata_required_exists
        or appointment.requisition_metadata_required_exists
    ):
        appointment.appt_status = INCOMPLETE_APPT
    else:
        appointment.appt_status = COMPLETE_APPT
    if save:
        appointment.save_base(update_fields=["appt_status"])
        appointment.refresh_from_db()
    return appointment


def get_previous_appointment(
    appointment: Appointment, include_interim: bool | None = None
) -> Appointment | None:
    """Returns the previous appointment model instance,
    or None, in this schedule.

    Keywords:
        * include_interim: include interim appointments
          (e.g. those where visit_code_sequence != 0)

    See also: `AppointmentMethodsModelMixin`
    """
    check_appointment_required_values_or_raise(appointment)
    opts: dict[str, str | int | Decimal] = dict(
        subject_identifier=appointment.subject_identifier,
        visit_schedule_name=appointment.visit_schedule_name,
        schedule_name=appointment.schedule_name,
    )
    if include_interim:
        if appointment.visit_code_sequence != 0:
            opts.update(
                visit_code_sequence__lt=appointment.visit_code_sequence,
                timepoint__lte=appointment.timepoint,
            )
        else:
            opts.update(timepoint__lt=appointment.timepoint)
    elif not include_interim:
        opts.update(
            visit_code_sequence=0,
            timepoint__lt=appointment.timepoint,
        )

    appointments: QuerySet[Appointment] = (
        appointment.__class__.objects.filter(**opts)
        .exclude(id=appointment.id)
        .order_by("timepoint", "visit_code_sequence")
    )

    try:
        previous_appt = appointments.reverse()[0]
    except IndexError:
        previous_appt = None
    return previous_appt


def get_next_appointment(appointment: Appointment, include_interim=None) -> Appointment | None:
    """Returns the next appointment model instance,
    or None, in this schedule.

    Keywords:
        * include_interim: include interim appointments
          (e.g. those where visit_code_sequence != 0)

    See also: `AppointmentMethodsModelMixin`
    """
    next_appt: Appointment | None = None
    check_appointment_required_values_or_raise(appointment)
    opts: dict[str, str | int | Decimal] = dict(
        subject_identifier=appointment.subject_identifier,
        visit_schedule_name=appointment.visit_schedule_name,
        schedule_name=appointment.schedule_name,
    )
    if include_interim:
        break_on_next = False
        for obj in appointment.__class__.objects.filter(
            timepoint__gte=appointment.timepoint, **opts
        ).order_by("timepoint", "visit_code_sequence"):
            if break_on_next:
                next_appt = obj
                break
            if obj.id == appointment.id:
                break_on_next = True
    elif not include_interim:
        opts.update(
            timepoint__gt=appointment.timepoint,
            visit_code_sequence=0,
        )
        next_appt = (
            appointment.__class__.objects.filter(**opts)
            .exclude(id=appointment.id)
            .order_by("timepoint", "visit_code_sequence")
        ).first()
    return next_appt


def check_appointment_required_values_or_raise(appointment: Appointment) -> None:
    if (
        not appointment.visit_schedule_name
        or not appointment.schedule_name
        or not appointment.visit_code
        or appointment.visit_code_sequence is None
        or appointment.timepoint is None
    ):
        raise AppointmentMissingValuesError(
            f"Appointment instance is missing required values. See {appointment}."
        )


def raise_on_appt_datetime_not_in_window(
    appointment: Appointment,
    appt_datetime: datetime | None = None,
    baseline_timepoint_datetime: datetime | None = None,
) -> None:
    if appointment.appt_status != CANCELLED_APPT and not is_baseline(instance=appointment):
        baseline_timepoint_datetime = baseline_timepoint_datetime or (
            appointment.__class__.objects.first_appointment(
                subject_identifier=appointment.subject_identifier,
                visit_schedule_name=appointment.visit_schedule_name,
                schedule_name=appointment.schedule_name,
            ).timepoint_datetime
        )
        try:
            appointment.schedule.datetime_in_window(
                dt=appt_datetime or appointment.appt_datetime,
                timepoint_datetime=appointment.timepoint_datetime,
                visit_code=appointment.visit_code,
                visit_code_sequence=appointment.visit_code_sequence,
                baseline_timepoint_datetime=baseline_timepoint_datetime,
            )
        except ScheduledVisitWindowError as e:
            msg = str(e)
            msg.replace("Invalid datetime", "Invalid appointment datetime (S)")
            raise AppointmentWindowError(msg) from e
        except UnScheduledVisitWindowError as e:
            msg = str(e)
            msg.replace("Invalid datetime", "Invalid appointment datetime (U)")
            raise AppointmentWindowError(msg) from e


def get_window_gap_days(appointment) -> int:
    """Return the number of days betwen this visit's upper and the
    next visit's lower.

    See get_default_max_visit_window_gap and settings attr.
    """
    if not appointment.next:
        gap_days = 0
    else:
        gap_days = abs(
            (appointment.timepoint_datetime + appointment.visit.rupper)
            - (appointment.next.timepoint_datetime - appointment.next.visit.rlower)
        ).days
    return gap_days


def appt_datetime_in_gap(appointment: Appointment, suggested_appt_datetime: datetime) -> bool:
    """Return True if datetime falls in a gap between this and the
    next appointment window.
    """
    in_gap = False
    if get_window_gap_days(appointment) > 0:
        next_lower_datetime = (
            appointment.next.timepoint_datetime - appointment.next.visit.rlower
        )
        upper_datetime = appointment.timepoint_datetime + appointment.visit.rupper
        if upper_datetime < suggested_appt_datetime < next_lower_datetime:
            in_gap = True
    return in_gap


def get_max_window_gap_to_lower(appointment) -> int:
    if (
        appointment.visit.max_window_gap_to_lower is not None
        and appointment.visit.max_window_gap_to_lower < get_default_max_visit_window_gap()
    ):
        max_gap = appointment.visit.max_window_gap_to_lower
    else:
        max_gap = get_default_max_visit_window_gap()
    return max_gap


def appt_datetime_in_next_window_adjusted_for_gap(
    appointment: Appointment, suggested_appt_datetime: datetime
) -> bool:
    """Returns True if `suggest_datetime` falls between the
    NEXT appointment's lower and upper window period datetime after
    adding gap_days to the lower datetime.
    """
    in_window = False
    gap_days = get_window_gap_days(appointment)
    max_gap = get_max_window_gap_to_lower(appointment)
    gap_days = min(gap_days, max_gap)
    if gap_days > 0:
        next_lower_datetime = (
            appointment.next.timepoint_datetime
            - appointment.next.visit.rlower
            - relativedelta(days=gap_days)
        )
        next_upper_datetime = (
            appointment.next.timepoint_datetime + appointment.next.visit.rupper
        )
        if next_lower_datetime <= suggested_appt_datetime <= next_upper_datetime:
            in_window = True
    return in_window


def get_appointment_by_datetime(
    suggested_appt_datetime: datetime,
    subject_identifier: str,
    visit_schedule_name: str,
    schedule_name: str,
    raise_if_in_gap: bool | None = None,
) -> Appointment | None:
    """Returns an appointment where the suggested datetime falls
    within the window period.

    * Returns None if no appointment is found.
    * Raises an exception if there is a gap between upper and lower
      boundaries and the date falls within the gap.
    """
    appointment = None
    raise_if_in_gap = True if raise_if_in_gap is None else raise_if_in_gap
    appointments = (
        django_apps.get_model("edc_appointment.appointment")
        .objects.filter(
            subject_identifier=subject_identifier,
            visit_schedule_name=visit_schedule_name,
            schedule_name=schedule_name,
            visit_code_sequence=0,
        )
        .order_by("timepoint_datetime")
    )
    for appointment in appointments:
        if appointment.appt_status == CANCELLED_APPT or is_baseline(appointment):
            continue
        try:
            raise_on_appt_datetime_not_in_window(
                appointment, appt_datetime=suggested_appt_datetime
            )
        except AppointmentWindowError as e:
            in_gap = appt_datetime_in_gap(appointment, suggested_appt_datetime)
            in_next_window_adjusted = appt_datetime_in_next_window_adjusted_for_gap(
                appointment, suggested_appt_datetime
            )
            if in_gap and raise_if_in_gap:
                dt = suggested_appt_datetime.strftime(
                    convert_php_dateformat(settings.SHORT_DATE_FORMAT)
                )
                raise AppointmentDateWindowPeriodGapError(
                    f"Date falls in a `window period gap` between {appointment.visit_code} "
                    f"and {appointment.next.visit_code}. Got {dt}."
                ) from e
            if (
                in_gap
                and in_next_window_adjusted
                and appointment.next.visit.add_window_gap_to_lower
            ):
                appointment = appointment.next
                break
            if (
                in_gap
                and not in_next_window_adjusted
                and appointment.next.visit.add_window_gap_to_lower
            ):
                appointment = None
                break
            appointment = appointment.next
        else:
            break
    return appointment


def reset_appointment(appointment: Appointment, **kwargs):
    """Reset appointment but only if appointment has not started.

    Will overwrite the default field values with values
    from kwargs.

    If field in kwargs refers to a field class with a related_model
    and the value is not found, the field will be set to None without
    raising an ObjectDoesNotExist exception.
    """
    if has_keyed_metadata(appointment) or appointment.related_visit:
        raise AppointmentAlreadyStarted(
            f"Unable to reset. Appointment already started. Got {appointment}."
        )
    defaults = dict(
        appt_status=appointment._meta.get_field("appt_status").default,
        appt_timing=appointment._meta.get_field("appt_timing").default,
        appt_type=None,
        appt_type_other="",
        appt_datetime=appointment.timepoint_datetime,
        comment="",
    )
    defaults.update(**kwargs)
    for k, v in defaults.items():
        try:
            related_model = get_model_from_relation(appointment._meta.get_field(k))
        except NotRelationField:
            setattr(appointment, k, v)
        else:
            try:
                setattr(appointment, k, related_model.objects.get(name=v))
            except ObjectDoesNotExist:
                setattr(appointment, k, None)
    appointment.save_base(update_fields=[*defaults.keys()])


def skip_appointment(appointment: Appointment, comment: str | None = None):
    """Set appointment to `SKIPPED_APPT` if appointment has not
    started.
    """
    if has_keyed_metadata(appointment) or appointment.related_visit:
        raise AppointmentAlreadyStarted(
            f"Unable to skip. Appointment already started. Got {appointment}."
        )
    reset_appointment(
        appointment,
        appt_status=SKIPPED_APPT,
        appt_timing=NOT_APPLICABLE,
        appt_type=NOT_APPLICABLE,
        comment=comment or "",
    )


def get_unscheduled_appointment_url(appointment: Appointment = None) -> str:
    """Returns a url for the unscheduled appointment."""
    dashboard_url_name = "subject_dashboard_url"
    dashboard_url = url_names.get(dashboard_url_name)
    unscheduled_appointment_url_name = "edc_appointment:unscheduled_appointment_url"
    kwargs = dict(
        subject_identifier=appointment.subject_identifier,
        visit_schedule_name=appointment.visit_schedule_name,
        schedule_name=appointment.schedule_name,
        visit_code=appointment.visit_code,
        visit_code_sequence=appointment.visit_code_sequence + 1,
        timepoint=appointment.timepoint,
    )
    kwargs.update(visit_code_sequence=str(appointment.visit_code_sequence + 1))
    kwargs.update(redirect_url=dashboard_url)
    return reverse(unscheduled_appointment_url_name, kwargs=kwargs)


def update_appt_status_for_timepoint(related_visit: RelatedVisitModel) -> None:
    """Only check COMPLETE_APPT and INCOMPLETE_APPT against metadata."""
    if related_visit.appointment.appt_status == COMPLETE_APPT:
        if (
            related_visit.metadata[CRF].filter(entry_status=REQUIRED).exists()
            or related_visit.metadata[REQUISITION].filter(entry_status=REQUIRED).exists()
        ):
            related_visit.appointment.appt_status = INCOMPLETE_APPT
            related_visit.appointment.save_base(update_fields=["appt_status"])
    elif related_visit.appointment.appt_status == INCOMPLETE_APPT and (
        not related_visit.metadata[CRF].filter(entry_status=REQUIRED).exists()
        and not related_visit.metadata[REQUISITION].filter(entry_status=REQUIRED).exists()
    ):
        related_visit.appointment.appt_status = COMPLETE_APPT
        related_visit.appointment.save_base(update_fields=["appt_status"])


def refresh_appointments(
    subject_identifier: str,
    visit_schedule_name: str,
    schedule_name: str,
    request: WSGIRequest | None = None,
    warn_only: bool | None = None,
) -> tuple[str, str]:
    status = OK
    visit_schedule = site_visit_schedules.get_visit_schedule(visit_schedule_name)
    schedule = visit_schedule.schedules.get(schedule_name)
    try:
        schedule.refresh_schedule(subject_identifier)
    except AppointmentDatetimeError as e:
        if request and not warn_only:
            status = ERROR_CODE
            messages.add_message(
                request=request,
                level=ERROR,
                message=_(
                    "An error was encountered when refreshing appointments. "
                    "Contact your administrator. Got '%(error_msg)s'."
                )
                % dict(error_msg=str(e)),
            )
        elif warn_only:
            warnings.warn(str(e), stacklevel=2)
        else:
            raise
    else:
        for appointment in get_appointment_model_cls().objects.filter(
            subject_identifier=subject_identifier,
            visit_schedule_name=visit_schedule_name,
            schedule_name=schedule_name,
        ):
            if appointment.related_visit:
                update_appt_status_for_timepoint(appointment.related_visit)
    if status == OK and request:
        messages.add_message(
            request,
            SUCCESS,
            _("The appointments for %(subject_identifier)s have been refreshed")
            % dict(subject_identifier=subject_identifier),
        )
    return subject_identifier, status


def validate_date_is_on_clinic_day(
    cleaned_data: dict = None, clinic_days=None, raise_validation_error: callable = None
):
    raise_validation_error = raise_validation_error or ValidationError
    if cleaned_data.get("appt_date"):
        try:
            appt_date = to_local(cleaned_data.get("appt_date")).date()
        except AttributeError:
            appt_date = cleaned_data.get("appt_date")
        try:
            report_date = to_local(cleaned_data.get("report_datetime")).date()
        except AttributeError:
            report_date = cleaned_data.get("report_datetime")
        day_abbr = calendar.weekheader(3).split(" ")
        if appt_date == report_date:
            raise raise_validation_error(
                {"appt_date": "Cannot be equal to the report datetime"}, INVALID_ERROR
            )
        if appt_date <= report_date:
            raise raise_validation_error(
                {"appt_date": "Cannot be before the report datetime"}, INVALID_ERROR
            )
        if not allow_clinic_on_weekend() and calendar.weekday(
            appt_date.year, appt_date.month, appt_date.day
        ) in [calendar.SATURDAY, calendar.SUNDAY]:
            raise raise_validation_error(
                {
                    "appt_date": _("Expected %(mon)s-%(fri)s. Got %(day)s")
                    % {
                        "mon": day_abbr[calendar.MONDAY],
                        "fri": day_abbr[calendar.FRIDAY],
                        "day": day_abbr[
                            calendar.weekday(appt_date.year, appt_date.month, appt_date.day)
                        ],
                    }
                },
                INVALID_ERROR,
            )
        if (
            clinic_days
            and calendar.weekday(appt_date.year, appt_date.month, appt_date.day)
            not in clinic_days
        ):
            days_str = [day_abbr[d] for d in clinic_days]
            days_str = []
            for d in clinic_days:
                days_str.append(day_abbr[d])
            raise raise_validation_error(
                {
                    "appt_date": _(
                        "Invalid clinic day. Expected %(expected)s. Got %(day_abbr)s"
                    )
                    % {
                        "expected": ", ".join(days_str),
                        "day_abbr": appt_date.strftime("%A"),
                    }
                },
                INVALID_ERROR,
            )



from django.urls import path
from django.views.generic import RedirectView

from .admin_site import edc_qareports_admin

app_name = "edc_qareports"

urlpatterns = [
    path("/admin/", edc_qareports_admin.urls),
    path("", RedirectView.as_view(url="/edc_qareports/admin/"), name="home_url"),
]



INSULIN = "insulin"
DIET_LIFESTYLE = "diet_lifestyle"
DRUGS = "drugs"
DIABETES_CLINIC = "diabetes_clinic"
HIV_CLINIC = "hiv_clinic"
HYPERTENSION_CLINIC = "hypertension_clinic"
THIS_CLINIC = "this_clinic"



from urllib.parse import unquote, urlencode

from django import template
from django.apps import apps as django_apps
from django.core.exceptions import ObjectDoesNotExist

register = template.Library()


@register.inclusion_tag("edc_offstudy/visit_schedule_row.html")
def offstudy_visit_schedule_row(subject_identifier, visit_schedule, subject_dashboard_url):
    context = {}
    offstudy_model = visit_schedule.offstudy_model
    offstudy_model_cls = django_apps.get_model(offstudy_model)
    try:
        obj = offstudy_model_cls.objects.get(subject_identifier=subject_identifier)
    except ObjectDoesNotExist:
        pass
    else:
        options = dict(subject_identifier=subject_identifier)
        query = unquote(urlencode(options))
        href = f"{obj.get_absolute_url()}?next={subject_dashboard_url},subject_identifier"
        href = "&".join([href, query])
        context = dict(
            offstudy_datetime=obj.offstudy_datetime,
            visit_schedule=visit_schedule,
            href=href,
            verbose_name=obj._meta.verbose_name,
        )
    return context



from .group_admin import GroupAdmin
from .log_entries_admin import LogEntriesAdmin
from .role_admin import RoleAdmin
from .user_admin import UserAdmin
from .user_profile_admin import UserProfileAdmin



from .administration_view_mixin import AdministrationViewMixin
from .edc_view_mixin import EdcViewMixin
from .message_view_mixin import MessageViewMixin
from .template_request_context_mixin import TemplateRequestContextMixin
from .url_request_context_mixin import UrlRequestContextMixin



# Generated by Django 2.1.2 on 2018-11-11 21:36

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [("edc_notification", "0001_initial")]

    operations = [
        migrations.AlterField(
            model_name="notification",
            name="name",
            field=models.CharField(max_length=255, unique=True),
        )
    ]



class PanelAlreadyRegistered(Exception):
    pass


class AddPanelError(Exception):
    pass


class LabProfileRequisitionModelError(Exception):
    pass


class LabProfile:
    """A container class for aliquot types, panels and processing."""

    site_model = "sites.site"

    def __init__(
        self, name=None, requisition_model=None, reference_range_collection_name=None
    ):
        self.aliquot_types = {}
        self.processing_profiles = {}
        self.panels = {}
        self.name = name
        if not requisition_model:
            raise LabProfileRequisitionModelError("Invalid requisition model. Got None")
        self.requisition_model = requisition_model
        # name for reportables collection, may also be set by RequisitionPanel
        self.reference_range_collection_name = reference_range_collection_name

    def __repr__(self):
        return f"{self.__class__.__name__}(name={self.name})"

    def __str__(self):
        return self.name

    def add_panel(self, panel):
        """Adds a requisition panel or reuqisition panel group
        instance to the profile.
        """

        panel.requisition_model = self.requisition_model
        panel.lab_profile_name = self.name
        if not panel.reference_range_collection_name:
            panel.reference_range_collection_name = self.reference_range_collection_name
        if panel.name in self.panels:
            raise PanelAlreadyRegistered(f"Panel already registered. Got {panel.name}")
        self.panels.update({panel.name: panel})
        self.processing_profiles.update(
            {panel.processing_profile.name: panel.processing_profile}
        )
        self.aliquot_types.update({panel.name: panel})
        self.aliquot_types.update({panel.numeric_code: panel})
        self.aliquot_types.update({panel.alpha_code: panel})



# Generated by Django 4.2.7 on 2023-12-04 02:32

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("edc_export", "0017_alter_datarequest_options_and_more"),
    ]

    operations = [
        migrations.RemoveIndex(
            model_name="objecthistory",
            name="edc_export__modifie_36e5ad_idx",
        ),
        migrations.RemoveIndex(
            model_name="plan",
            name="edc_export__modifie_db36ec_idx",
        ),
        migrations.AddIndex(
            model_name="objecthistory",
            index=models.Index(
                fields=["modified", "created"], name="edc_export__modifie_808a80_idx"
            ),
        ),
        migrations.AddIndex(
            model_name="plan",
            index=models.Index(
                fields=["modified", "created"], name="edc_export__modifie_25ba96_idx"
            ),
        ),
    ]



import sys

from django.conf import settings
from django.core.management.color import color_style

style = color_style()


def populate_visit_schedule(sender=None, **kwargs):
    from .models import VisitSchedule  # noqa: PLC0415
    from .site_visit_schedules import site_visit_schedules  # noqa: PLC0415

    sys.stdout.write(style.MIGRATE_HEADING("Populating visit schedule:\n"))
    if getattr(settings, "EDC_VISIT_SCHEDULE_POPULATE_VISIT_SCHEDULE", True):
        VisitSchedule.objects.update(active=False)
        site_visit_schedules.to_model(VisitSchedule)
        sys.stdout.write("Done.\n")
    else:
        sys.stdout.write(
            "  not populating. See settings."
            "EDC_VISIT_SCHEDULE_POPULATE_VISIT_SCHEDULE. Done.\n"
        )
    sys.stdout.flush()



from django.conf import settings


def get_autodiscover():
    return getattr(settings, "EDC_NAVBAR_AUTODISCOVER", True)


def get_verify_on_load():
    return getattr(settings, "EDC_NAVBAR_VERIFY_ON_LOAD", "")



from __future__ import annotations

from django import template
from django.urls import reverse
from django.urls.exceptions import NoReverseMatch

from ..navbar_item import NavbarItem

register = template.Library()


@register.inclusion_tag("edc_navbar/edc_navbar.html", takes_context=True)
def render_navbar(context) -> dict:
    auth_user_change_url = None
    user = getattr(context["request"], "user", None)
    user_id = getattr(user, "id", None)
    try:
        auth_user_change_url = reverse("edc_auth_admin:auth_user_change", args=(user_id,))
    except NoReverseMatch:
        pass
    return dict(
        auth_user_change_url=auth_user_change_url,
        default_navbar=context.get("default_navbar"),
        navbar=context.get("navbar"),
        user=user,
        request=context["request"],
    )


@register.inclusion_tag("edc_navbar/navbar_item.html", takes_context=True)
def render_navbar_item(context, navbar_item: NavbarItem) -> dict:
    data = {}
    url = navbar_item.get_url()
    navbar_item.set_disabled(user=getattr(context["request"], "user", None))
    if url:
        data.update(url=url)
    data.update(navbar_item=navbar_item)
    return data



from .research_protocol_config import ResearchProtocolConfig


class ResearchProtocolConfigMiddleware:
    def __init__(self, get_response):
        self.get_response = get_response

    def __call__(self, request):
        return self.get_response(request)

    def process_view(self, request, *args):
        pass

    def process_template_response(self, request, response):
        if not response.context_data:
            response.context_data = {}
        protocol_config = ResearchProtocolConfig()
        response.context_data.update(
            copyright=protocol_config.copyright,
            disclaimer=protocol_config.disclaimer,
            institution=protocol_config.institution,
            license=protocol_config.license,
            project_name=protocol_config.project_name,
        )
        return response



import re

from django.conf import settings
from django.core.exceptions import ValidationError


def phone_number(value, pattern=None):
    str_value = str(value)
    pattern = pattern or r"^[0-9+\(\)#\.\s\/ext-]+$"
    p = re.compile(pattern)
    if not p.match(str_value):
        raise ValidationError("Invalid format.")


def cell_number(value):
    try:
        pattern = settings.CELLPHONE_REGEX
    except AttributeError:
        pattern = None
    phone_number(value, pattern)


def telephone_number(value):
    try:
        pattern = settings.TELEPHONE_REGEX
    except AttributeError:
        pattern = None
    phone_number(value, pattern)



import csv
from pathlib import Path

from django.apps import apps as django_apps
from django.utils import timezone


class FileHistoryUpdater:
    file_history_model = "edc_export.filehistory"

    def __init__(
        self,
        path: Path,
        delimiter: str,
        model: str,
        filename: str,
        notification_plan_name: str | None = None,
    ):
        self.model = model
        self.filename = filename
        self.notification_plan_name = notification_plan_name or "Notification plan"
        self.path = path
        self.delimiter = delimiter

    @property
    def model_cls(self):
        return django_apps.get_model(self.file_history_model)

    def update(self):
        exported_pks = []
        export_uuids = []
        with self.path.open("r") as f:
            csv_reader = csv.DictReader(f, delimiter=self.delimiter)
            for row in csv_reader:
                exported_pks.append(row["id"])
                export_uuids.append(row["export_uuid"])
        return self.model_cls.objects.create(
            model=self.model,
            pk_list="|".join(exported_pks),
            export_uuid_list="|".join(export_uuids),
            exported=True,
            exported_datetime=timezone.now(),
            filename=self.filename,
            notification_plan_name=self.notification_plan_name,
        )



from __future__ import annotations

import json
from typing import TYPE_CHECKING

from django.core.exceptions import ObjectDoesNotExist

if TYPE_CHECKING:
    from ..models import SiteProfile


def get_or_create_site_profile_obj(single_site, site_obj, apps) -> SiteProfile | None:
    site_profile_model_cls = apps.get_model("edc_sites", "SiteProfile")
    opts = dict(
        title=single_site.description,
        country=single_site.country,
        country_code=single_site.country_code,
        languages=json.dumps(single_site.languages) if single_site.languages else None,
    )
    try:
        site_profile = site_profile_model_cls.objects.get(site=site_obj)
    except ObjectDoesNotExist:
        site_profile = site_profile_model_cls.objects.create(site=site_obj, **opts)
    else:
        for k, v in opts.items():
            setattr(site_profile, k, v)
        site_profile.save()
    return site_profile



from django.apps import apps as django_apps

from edc_lab.utils import get_requisition_model_name

LAB = "LAB"
LAB_VIEW = "LAB_VIEW"
LAB_TECHNICIAN_ROLE = "laboratory_technician"

lab_requisition = []
for action in ["view_", "add_", "change_", "delete_", "view_historical"]:
    lab_requisition.append(f".{action}".join(get_requisition_model_name().split(".")))


lab_codenames = []
for app_config in django_apps.get_app_configs():
    if app_config.name in ["edc_lab"]:
        for model_cls in app_config.get_models():
            for prefix in ["add", "change", "delete", "view"]:
                lab_codenames.append(
                    f"{app_config.name}.{prefix}_{model_cls._meta.model_name}"
                )
lab_codenames.sort()


lab_codenames.extend(lab_requisition)


lab_view_codenames = [
    c for c in lab_codenames if ("view_" in c or "edc_nav" in c or "edc_dashboard" in c)
]



# Generated by Django 2.0 on 2017-12-30 13:41

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [("edc_lab", "0010_auto_20171127_1541")]

    operations = [migrations.DeleteModel(name="IdentifierHistory")]



from .action_views import (
    AliquotView,
    ManageBoxItemView,
    ManageManifestView,
    ManifestView,
    PackView,
    ProcessView,
    ReceiveView,
    RequisitionView,
    VerifyBoxItemView,
)
from .home_view import HomeView
from .listboard_views import (
    AliquotListboardView,
    ManageBoxListboardView,
    ManageManifestListboardView,
    ManifestListboardView,
    PackListboardView,
    ProcessListboardView,
    ReceiveListboardView,
    RequisitionListboardView,
    ResultListboardView,
    VerifyBoxListboardView,
)



from edc_list_data.model_mixins import ListModelMixin


class UnitType(ListModelMixin):
    """Unit type in a container.

    For example, `tablet`.
    """

    class Meta(ListModelMixin.Meta):
        pass



from .constants import (
    ALREADY_REPORTED,
    MILD,
    MODERATE,
    PRESENT_AT_BASELINE,
    SEVERE,
    SEVERITY_INCREASED_FROM_G3,
)
from .exceptions import (
    AlreadyRegistered,
    ConversionNotHandled,
    FormulaError,
    GradeReferenceError,
    LimitsNormalParseError,
)
from .formula import Formula, dummy_formula, formula
from .units import (
    CELLS_PER_MICROLITER,
    CELLS_PER_MILLIMETER_CUBED,
    CELLS_PER_MILLIMETER_CUBED_DISPLAY,
    COPIES_PER_MILLILITER,
    EGFR_UNITS,
    GRAMS_PER_DECILITER,
    GRAMS_PER_LITER,
    IU_LITER,
    IU_LITER_DISPLAY,
    MICROMOLES_PER_LITER,
    MICROMOLES_PER_LITER_DISPLAY,
    MILLIGRAMS_PER_DECILITER,
    MILLIGRAMS_PER_LITER,
    MILLILITER_PER_MINUTE,
    MILLIMOLES_PER_LITER,
    MILLIMOLES_PER_LITER_DISPLAY,
    MM3,
    MM3_DISPLAY,
    PERCENT,
    PLUS,
    TEN_X_3_PER_LITER,
    TEN_X_3_PER_LITER_DISPLAY,
    TEN_X_9_PER_LITER,
    TEN_X_9_PER_LITER_DISPLAY,
)



import numpy as np
import pandas as pd


def convert_numbers_to_nullable_dtype(df: pd.DataFrame) -> pd.DataFrame:
    for col in df.select_dtypes(include=["number"]).columns:
        if pd.api.types.is_float_dtype(df[col]):
            df[col] = df[col].astype("Float64")
        elif pd.api.types.is_integer_dtype(df[col]):
            df[col] = df[col].astype("Int64")
    df.replace({np.nan: pd.NA})
    return df



# from .crf_metadata_wrapper import CrfMetadataWrapper
# from .crf_metadata_wrappers import CrfMetadataWrappers
# from .metadata_wrapper import MetadataWrapper, MetadataWrapperError
# from .requisition_metadata_wrapper import RequisitionMetadataWrapper
# from .requisition_metadata_wrappers import RequisitionMetadataWrappers



from __future__ import annotations

from datetime import datetime
from decimal import Decimal
from typing import TYPE_CHECKING

from tqdm import tqdm

from edc_registration.models import RegisteredSubject

from ..site_data_manager import site_data_manager

if TYPE_CHECKING:
    from ..models import QueryRule, QueryVisitSchedule


class QueryRuleWrapper:
    def __init__(
        self,
        query_rule_obj: QueryRule | None = None,
        subject_identifiers: list[str] | None = None,
        visit_schedule_obj: QueryVisitSchedule | None = None,
        timepoint: Decimal | None = None,
        entry_status: str | None = None,
        now: datetime | None = None,
        verbose: bool | None = None,
    ):
        self.now = now
        self.query_rule_obj = query_rule_obj
        self.subject_identifiers = subject_identifiers
        self.visit_schedule_obj = visit_schedule_obj
        self.timepoint = timepoint
        self.entry_status = entry_status
        self.verbose = verbose

    def __repr__(self):
        return (
            f"{self.__class__.__name__}("
            f"model={self.query_rule_obj.model_cls._meta.label_lower}, "
            f"query_rule={self.query_rule_obj.title}, timepoint={self.timepoint}, "
            f"{self.entry_status}, "
            f"subject_identifiers={self.count(self.subject_identifiers)})"
        )

    def __str__(self):
        return (
            f"model={self.query_rule_obj.model_cls._meta.label_lower}, "
            f"query_rule={self.query_rule_obj.title}, timepoint={self.timepoint}, "
            f"{self.entry_status}, "
            f"subject_identifiers={self.count(self.subject_identifiers)})"
        )

    @property
    def handler(self):
        return site_data_manager.get_rule_handler(self.query_rule_obj.rule_handler_name)

    def run_handler(self):
        """Run handler for each subject_identifier / visit_schedule / timepoint.

        Will create data queries if needed.
        """
        created = 0
        resolved = 0
        total = self.count(self.subject_identifiers)
        for value in tqdm(
            self.get_subject_identifiers(), total=total, disable=not self.verbose
        ):
            registered_subject = RegisteredSubject.objects.get(
                subject_identifier=value["subject_identifier"]
            )
            handler = self.handler(
                query_rule_obj=self.query_rule_obj,
                registered_subject=registered_subject,
                visit_schedule_obj=self.visit_schedule_obj,
                now=self.now,
            )
            handler.run()
            created += handler.created_counter
            resolved += handler.resolved_counter
        return created, resolved

    @property
    def get_subject_identifiers(self):
        """Returns an iterable of dictionaries of
        {subject_identifier: value, ...}.
        """
        return getattr(
            self.subject_identifiers,
            "all",
            lambda: [{"subject_identifier": s for s in self.subject_identifiers}],
        )

    @staticmethod
    def count(qs_or_list):
        try:
            cnt = qs_or_list.count()
        except (TypeError, AttributeError):
            cnt = len(qs_or_list)
        return cnt



# Generated by Django 4.2.7 on 2023-12-04 22:19

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        (
            "edc_facility",
            "0010_remove_healthfacility_edc_facilit_modifie_dcf8ac_idx_and_more",
        ),
    ]

    operations = [
        migrations.RemoveIndex(
            model_name="healthfacilitytypes",
            name="edc_facilit_name_78007a_idx",
        ),
    ]



from django.apps import apps as django_apps
from django.contrib import messages
from django.utils import timezone

from edc_constants.constants import YES
from edc_dashboard.view_mixins import EdcViewMixin
from edc_lab import Specimen
from edc_lab.labels import AliquotLabel
from edc_lab.site_labs import site_labs

from ...view_mixins import ProcessRequisitionViewMixin
from .action_view import ActionView


class ReceiveView(EdcViewMixin, ProcessRequisitionViewMixin, ActionView):
    post_action_url = "receive_listboard_url"
    valid_form_actions = ["receive", "receive_and_process"]
    label_cls = AliquotLabel
    specimen_cls = Specimen

    def process_form_action(self, request=None):
        if not self.selected_items:
            message = "Nothing to do. No items selected."
            messages.warning(self.request, message)
        if self.action == "receive":
            self.receive()
            self.create_specimens()
        elif self.action == "receive_and_process":
            self.receive()
            self.create_specimens()
            self.process(request)

    def receive(self):
        """Updates selected requisitions as received."""
        updated = 0
        for model in site_labs.requisition_models.values():
            model_cls = django_apps.get_model(model)
            updated += (
                model_cls.objects.filter(pk__in=self.selected_items, is_drawn=YES)
                .exclude(received=True)
                .update(received=True, received_datetime=timezone.now())
            )
        if updated:
            message = f"{updated} requisitions received."
            messages.success(self.request, message)
        return updated

    def create_specimens(self):
        """Creates aliquots for each selected and received requisition."""
        for requisition in self.get_requisitions(pk__in=self.selected_items, received=True):
            self.specimen_cls(requisition=requisition)



from django.conf import settings
from django.template.loader import render_to_string
from django.utils import timezone

from edc_action_item.action import Action
from edc_action_item.site_action_items import site_action_items
from edc_constants.constants import (
    CLOSED,
    FEEDBACK,
    HIGH_PRIORITY,
    MEDIUM_PRIORITY,
    NORMAL,
    RESOLVED,
)

from .constants import CLOSED_WITH_ACTION

DATA_QUERY_ACTION = "data_query_action"
SHOW_ON_DASHBOARD = getattr(settings, "DATA_MANAGER_SHOW_ON_DASHBOARD", True)


class DataQueryAction(Action):
    name = DATA_QUERY_ACTION
    display_name = "Data query"
    reference_model = "edc_data_manager.dataquery"
    priority = MEDIUM_PRIORITY
    create_by_user = True
    show_link_to_changelist = True
    show_on_dashboard = SHOW_ON_DASHBOARD
    admin_site_name = "edc_data_manager_admin"
    instructions = "Review and respond to the query"
    delete_with_reference_object = True

    def reopen_action_item_on_change(self):
        return not self.close_action_item_on_save()

    def close_action_item_on_save(self):
        return self.reference_obj and self.reference_obj.status in [
            CLOSED,
            CLOSED_WITH_ACTION,
        ]

    def get_priority(self):
        return self.query_priority or self.priority

    def get_popover_title(self):
        return self.get_category()

    @property
    def site_response_status(self):
        try:
            site_response_status = self.reference_obj.site_response_status
        except AttributeError:
            site_response_status = None
        return site_response_status

    @property
    def query_priority(self):
        try:
            query_priority = self.reference_obj.query_priority
        except AttributeError:
            query_priority = None
        return query_priority

    def get_color_style(self):
        color_style = "warning"
        if self.site_response_status in [FEEDBACK, RESOLVED]:
            color_style = "info"
        return color_style

    def get_category(self):
        return (
            "DM Query" if self.site_response_status in [FEEDBACK, RESOLVED] else "Site Query"
        )

    def get_status(self):
        if self.reference_obj:
            return self.reference_obj.get_site_response_status_display()
        return "New"

    def get_display_name(self):
        template_name = "edc_data_manager/action_item_display_name.html"
        title = getattr(self.reference_obj, "title", "")
        visit_schedule = getattr(self.reference_obj, "visit_schedule", "")
        modified = getattr(self.reference_obj, "modified", None)
        auto = "auto" if getattr(self.reference_obj, "rule_generated", False) else ""
        try:
            query_priority = self.reference_obj.get_query_priority_display()
        except AttributeError:
            query_priority = NORMAL
        context = dict(
            HIGH_PRIORITY=HIGH_PRIORITY,
            auto=auto,
            category=self.get_category(),
            modified=modified,
            object=self.reference_obj,
            query_priority=query_priority,
            title=title,
            utcnow=timezone.now(),
            visit_schedule=visit_schedule,
        )
        form_and_numbers_to_string = getattr(
            self.reference_obj, "form_and_numbers_to_string", None
        )
        if form_and_numbers_to_string:
            context.update(form_and_numbers=form_and_numbers_to_string())
        return render_to_string(template_name, context=context)


site_action_items.register(DataQueryAction)



# Generated by Django 2.2.7 on 2019-12-13 00:14

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("edc_data_manager", "0013_auto_20191120_0304"),
    ]

    operations = [
        migrations.AlterField(
            model_name="historicalqueryrule",
            name="rule_handler_name",
            field=models.CharField(
                choices=[("do_nothing", "Do Nothing"), ("default", "Default")],
                default="default",
                max_length=150,
            ),
        ),
        migrations.AlterField(
            model_name="queryrule",
            name="rule_handler_name",
            field=models.CharField(
                choices=[("do_nothing", "Do Nothing"), ("default", "Default")],
                default="default",
                max_length=150,
            ),
        ),
    ]



from dataclasses import dataclass

from .crf_case import CrfCase
from .requisition_subquery import RequisitionSubquery


@dataclass(kw_only=True)
class RequisitionCase(CrfCase):
    panel: str = None
    subjectrequisition_dbtable: str | None = None
    panel_dbtable: str | None = None
    requisition_id_field: str | None = None

    @property
    def sql(self):
        sql = RequisitionSubquery(**self.__dict__).sql
        return sql.format(panel=self.panel)



# Generated by Django 2.2.2 on 2019-06-29 16:00

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [("edc_visit_schedule", "0003_historicalvisitschedule_visitschedule")]

    operations = [
        migrations.AddIndex(
            model_name="visitschedule",
            index=models.Index(
                fields=[
                    "visit_schedule_name",
                    "schedule_name",
                    "visit_code",
                    "visit_name",
                    "visit_title",
                ],
                name="edc_visit_s_visit_s_a90951_idx",
            ),
        )
    ]
